from copy import deepcopy
import tempfile
from typing import List
from icolos.core.containers.compound import Conformer, Enumeration
from icolos.core.workflow_steps.step import StepBase
from pydantic import BaseModel

try:
    from espsim import GetEspSim, GetShapeSim
except ImportError:
    print(
        "WARNING - Could not import module espsim, check it is installed in your environment"
    )

from rdkit.Chem import AllChem, Mol
from rdkit import Chem
from rdkit.Chem import rdFMCS
from rdkit.Chem import AllChem
from icolos.core.workflow_steps.step import _LE
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer
import os

# Based on https://github.com/hesther/espsim


class StepEspSim(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def _compute_esp_sim(self, ref: Mol, trg: Enumeration, tmp_dir: str):
        """
        :param ref : Reference molecule, the known binder against which to calculate similarity
        :param trg: Icolos enumeration of the target molecule, as a smiles string.  Embedded with RDKit
        """
        # Create mol object from trg smile string
        # housekeeping for data appending later

        trg_mol = Chem.AddHs(Chem.MolFromSmiles(trg.get_smile()))
        AllChem.EmbedMolecule(ref, AllChem.ETKDG())
        AllChem.EmbedMolecule(trg_mol, AllChem.ETKDG())
        mols = [Chem.RemoveHs(x) for x in [ref, trg_mol]]
        mcs = rdFMCS.FindMCS(
            mols,
            threshold=0.8,
            completeRingsOnly=True,
            ringMatchesRingOnly=True,
        )
        self._logger.log(
            f"Computed mcs: {mcs} for enumeration {trg.get_smile()}", _LE.DEBUG
        )

        patt = Chem.MolFromSmarts(mcs.smartsString)
        refMol = mols[0]
        refMatch = refMol.GetSubstructMatch(patt)

        mv = trg_mol.GetSubstructMatch(patt)
        AllChem.AlignMol(trg_mol, refMol, atomMap=list(zip(mv, refMatch)))

        # Generate an single conformer for the reference

        # All atom alignment of the two

        charge_method = self._get_additional_setting("charge_method", default="am1-bcc")
        esp_sim = GetEspSim(trg_mol, ref, partialCharges=charge_method)
        shape_sim = GetShapeSim(trg_mol, ref)
        self._logger.log(
            f"Computed EspSim: {esp_sim}, ShapeSim: {shape_sim} for mol {trg.get_smile()}",
            _LE.DEBUG,
        )

        # now attach the mols as conformers attach the scores to the mol objects
        trg_conf = Conformer(conformer=trg_mol)
        trg_conf.get_molecule().SetProp("shape_sim", str(shape_sim))
        trg_conf.get_molecule().SetProp("esp_sim", str(esp_sim))

        trg_conf.write(os.path.join(tmp_dir, "conformer.sdf"))

    def _get_arguments(self, std_args: List) -> List:

        for flag in self.settings.arguments.flags:
            std_args.append(flag)
        for key, value in self.settings.arguments.parameters.items():
            std_args.append(f"{key}=")
            std_args.append(value)
        return std_args

    def _prepare_batch(self, batch):
        target_enums = []
        tmp_dirs = []

        for sublist in batch:
            for task in sublist:
                target_enums.append(task.data)
                tmp_dirs.append(tempfile.mkdtemp())
        return target_enums, tmp_dirs

    def _parse_output(self, trgs: List[Enumeration], tmp_dirs: List[str]) -> None:
        for tmp_dir, trg in zip(tmp_dirs, trgs):
            # grab the written sdf object
            sdf_path = os.path.join(tmp_dir, "conformer.sdf")
            mol_supplier = Chem.SDMolSupplier(sdf_path, removeHs=False)
            for mol in mol_supplier:  # should only be one conformer!
                conf = Conformer(conformer=mol)
                comp = self.get_compound_by_name(trg.get_compound_name())
                comp.find_enumeration(trg.get_enumeration_id()).add_conformer(conf)

    def _execute_espsim_parallel(self):
        # embed the reference compound

        parallelizer = Parallelizer(func=self._compute_esp_sim)
        ref_compound = Chem.AddHs(
            Chem.MolFromSmiles(self.settings.additional["ref_smiles"])
        )
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())

            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            trgs, tmp_dirs = self._prepare_batch(next_batch)

            refs = [ref_compound for _ in range(len(next_batch))]

            parallelizer.execute_parallel(ref=refs, trg=trgs, tmp_dir=tmp_dirs)
            # hand over the embedded reference (compute once) and target compound (smiles string to be embedded)
            self._parse_output(tmp_dirs=tmp_dirs, trgs=trgs)

            for task, tmp_dir in zip(next_batch, tmp_dirs):
                for subtask in task:
                    if os.path.isfile(os.path.join(tmp_dir, "conformer.sdf")):
                        subtask.set_status_success()
                    else:
                        subtask.set_status_failed()

            self._remove_temporary(tmp_dirs)

    def execute(self):
        """
        esp-sim does molecular alignment with RDkit, then computes coulombic overlap integral + tanimoto similarity for shape measurement

        Use case takes a reference compound (known binder) and compare to REINVENT compounds

        Usage:
        * Define reference compound using settings.additional, as a smile string, to be embedded by RDkit
        * The remaining compounds are embedded using a preceeding RDkit embedding
        * attach the resulting scores to the enumeration
        """

        all_enums = []
        for compound in self.get_compounds():
            for enumeration in compound:
                all_enums.append(deepcopy(enumeration))
        if self._get_additional_setting("charge_method", default="am1-bcc") == "resp":
            # resp doesn#t play well with Icolos parallelization
            ref_compound = Chem.AddHs(
                Chem.MolFromSmiles(self.settings.additional["ref_smiles"])
            )
            tmp_dirs = []
            for enum in all_enums:
                tmp_dir = tempfile.mkdtemp()
                tmp_dirs.append(tmp_dir)
                self._compute_esp_sim(ref_compound, enum, tmp_dir=tmp_dir)
            self._parse_output(all_enums, tmp_dirs=tmp_dirs)
            self._remove_temporary(tmp_dirs)
        else:
            self.execution.parallelization.max_length_sublists = 1
            # unroll the provided compounds,
            self._subtask_container = SubtaskContainer(max_tries=3)
            self._subtask_container.load_data(all_enums)
            self._execute_espsim_parallel()



from typing import Dict, List

from pydantic import BaseModel, PrivateAttr
from icolos.core.containers.perturbation_map import PerturbationMap
from icolos.core.flow_control.flow_control import FlowControlBase
from icolos.core.workflow_steps.step import StepBase
from icolos.core.composite_agents.base_agent import BaseAgent, AgentHeaderParameters
from icolos.utils.enums.step_enums import StepBaseEnum

from icolos.utils.general.icolos_exceptions import get_exception_message

from icolos.utils.enums.logging_enums import LoggingConfigEnum

_LE = LoggingConfigEnum()
_SBE = StepBaseEnum


class WorkflowHeaderParameters(AgentHeaderParameters, BaseModel):
    pass


class WorkflowData(BaseModel):
    work_dir: str = None
    perturbation_map: PerturbationMap = None


class WorkFlow(BaseAgent, BaseModel):
    """Class to hold the whole logic for a workflow."""

    steps: List[Dict] = []
    header: WorkflowHeaderParameters = WorkflowHeaderParameters()
    workflow_data: WorkflowData = WorkflowData()

    class Config:
        underscore_attrs_are_private = True

    _logger = PrivateAttr()
    _initialized_steps = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._initialized_steps = []

    def initialize(self):
        from icolos.core.steps_utils import initialize_step_from_dict

        super().initialize()
        self._initialized_steps = []
        for step_conf in self.steps:
            step_conf = self._update_global_variables(conf=step_conf)
            step = initialize_step_from_dict(step_conf=step_conf)
            if isinstance(step, StepBase):
                # we have a normal step, no flow control wrapping
                step.set_workflow_object(self)
                self._initialized_steps.append(step)
            elif isinstance(step, FlowControlBase):
                self._initialized_steps.append(step.dispatcher)
                step.dispatcher.set_workflow_object(self)
        self._logger.log(
            f"Initialized {len(self._initialized_steps)} steps in workflow {self.header.id}.",
            _LE.DEBUG,
        )

    def execute(self):
        for step in self._initialized_steps:
            step.generate_input()
            self._logger.log(f"Starting execution of step: {step.step_id}", _LE.INFO)
            step.execute()
            self._logger.log(
                f"Processing write-out blocks for {step.step_id}.", _LE.DEBUG
            )
            step.process_write_out()
        self._logger.log(
            f"Execution of {len(self._initialized_steps)} steps completed.", _LE.INFO
        )

    def is_valid(self) -> bool:
        try:
            for step in self._initialized_steps:
                step.validate()
        except Exception as e:
            self._logger.log(
                f'During step validation, "WorkFlow" returned the following exception: {get_exception_message(e)}.',
                _LE.WARNING,
            )
            return False
        return True

    def add_step(self, step: StepBase):
        self._initialized_steps.append(step)

    def get_steps(self) -> list:
        return self._initialized_steps

    def find_step_by_step_id(self, step_id: str):
        for step in self._initialized_steps:
            if step.step_id == step_id:
                return step
            elif step.step_id == _SBE.STEP_DISPATCHER:
                pass

        raise IndexError(f"Could not find step with step_id {step_id} in workflow.")

    def __iter__(self):
        return iter(self.steps)

    def __repr__(self):
        return "<Icolos workflow: id=%s, description=%s, number steps: %s>" % (
            self.get_id(),
            self.get_description(),
            len(self),
        )

    def set_perturbation_map(self, p_map: PerturbationMap) -> None:
        self.workflow_data.perturbation_map = p_map

    def get_perturbation_map(self) -> PerturbationMap:
        return self.workflow_data.perturbation_map

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key: int):
        return self._initialized_steps[key]

    def __len__(self) -> int:
        return len(self._initialized_steps)



from icolos.core.containers.gmx_state import GromacsState
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE

_GE = GromacsEnum()
_SGE = StepGromacsEnum()


class StepGMXGrompp(StepGromacsBase, BaseModel):
    """
    Wraps gromacs preprocessor, produces tpr file preceeding mdrun step
    Automatically handles coupling group updates
    """

    class Config:
        arbitrary_types_allowed = True

    topol: GromacsState = None

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def _auto_update_coupling_groups(self, tmp_dir):
        # this will handle most straightforward cases with protein+ligand, DNA, RNA,
        result = self._generate_index_groups(tmp_dir)
        add_other = False
        add_ions = False

        # check whether the ions and other index groups are present
        for line in result.stdout.split("\n"):
            parts = line.split()

            if len(parts) == 5:
                if parts[1] in _GE.PRIMARY_COMPONENTS:
                    primary_component = parts[1]
                    # identify Protein, DNA, RNA
                elif parts[1] == "Other":
                    add_other = True
                elif parts[1] == _SGE.WATER_AND_IONS:
                    add_ions = True

        update_dict = self._copy_fields_dict()
        pipe_input = ""
        tc_grps = ""
        if add_other:
            pipe_input += f"{primary_component} or Other"
            tc_grps += f"{primary_component}_Other"
        else:
            tc_grps += primary_component
        if add_ions:
            tc_grps += " "
            tc_grps += _SGE.WATER_AND_IONS
        else:
            tc_grps += " "
            tc_grps += "Water"

        update_dict[_SGE.TC_GRPS] = tc_grps

        if pipe_input:
            self._add_index_group(tmp_dir, pipe_input)

        # update the mdp file with the modified coupling groups
        self._modify_config_file(
            tmp_dir,
            self.data.generic.get_argument_by_extension(
                _SGE.FIELD_KEY_MDP, rtn_file_object=True
            ),
            update_dict,
        )

    def execute(self):
        """
        Set up required mdp file and run gmx grompp
        Note that any issues with your parametrisation or system building will normally cause grompp to panic
        """
        tmp_dir = self._make_tmpdir()
        topol = self.get_topol()

        # do this bit once
        # if make_ndx command has been specified in settings.additional,
        # add an index group here, commonly protein_ligand or protein_other

        if (
            _SGE.MAKE_NDX_COMMAND in self.settings.additional.keys()
            and self.settings.additional[_SGE.MAKE_NDX_COMMAND] is not None
        ):
            # normally you want your two t-coupling groups to be something like Protein_Other Water_Ions
            # these can be added automatically with the "auto" keyword
            if self.settings.additional[_SGE.MAKE_NDX_COMMAND] == _SGE.AUTO:
                # automatically update the coupling groups, check for presence of 'ions' and 'other',
                # update default coupling groups in mdp file
                self._auto_update_coupling_groups(tmp_dir)
            else:
                # the mdp file will not be modified, coupling groups must be set correctly prior to job execution
                self._add_index_group(
                    tmp_dir, self.settings.additional[_SGE.MAKE_NDX_COMMAND]
                )

        mdp_file = self.data.generic.get_argument_by_extension(
            _SGE.FIELD_KEY_MDP, rtn_file_object=True
        )
        mdp_file.write(tmp_dir)

        replicas = self._get_additional_setting(_SGE.REPLICAS, default=1)
        # replicas = len(topol.tprs.values())
        topol.write_topol(tmp_dir)
        init_struct = len(topol.structures)

        for i in range(replicas):
            index = 0 if init_struct == 1 else i
            # we are branching for the first time, just use this confout.gro as the starting point
            topol.write_structure(tmp_dir, index=index)
            args = (
                ["-r", _SGE.STD_STRUCTURE]
                if self.settings.additional[_SGE.RESTRAINTS]
                else []
            )

            arguments = self._parse_arguments(
                flag_dict={
                    "-f": mdp_file.get_file_name(),
                    "-c": _SGE.STD_STRUCTURE,
                    "-p": _SGE.STD_TOPOL,
                    "-o": _SGE.STD_TPR,
                },
                args=args,
            )
            result = self._backend_executor.execute(
                command=_GE.GROMPP, arguments=arguments, check=True, location=tmp_dir
            )
            for line in result.stdout.split("\n"):
                self._logger_blank.log(line, _LE.DEBUG)
            self._logger.log(
                f"Completed execution for {self.step_id} successfully for replica {i}",
                _LE.INFO,
            )
            topol.set_tpr(tmp_dir, index=i)
            # topol.set_structure(tmp_dir, index=i)
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



from typing import Dict
from pydantic import BaseModel
import os
import sys

from icolos.core.composite_agents.workflow import WorkFlow
from icolos.utils.enums.logging_enums import LoggingConfigEnum

from icolos.utils.entry_point_functions.logging_helper_functions import (
    initialize_logging,
)
from icolos.utils.entry_point_functions.parsing_functions import (
    get_runtime_global_variables,
    add_global,
)
from icolos.utils.general.files_paths import attach_root_path
from icolos.utils.enums.composite_agents_enums import WorkflowEnum

_WE = WorkflowEnum()
_LE = LoggingConfigEnum()


class IcolosWorkflow(BaseModel):
    """
    Alternative programmatic entrypoint for the Icolos workflow
    """

    config: Dict = None
    workflow: WorkFlow = None
    logging: str = None
    global_vars: Dict = None

    def __init__(self, config, global_vars: Dict = None) -> None:
        super().__init__(**config)

        self.config = self._parse_global_vars(config, global_vars)
        self.workflow = WorkFlow(**config[_WE.WORKFLOW])
        # tutorial settings logs everything to stdout as well as the file
        self.logging = "tutorial"

    def _initialize_logging(self):
        log_conf = attach_root_path(_LE.PATH_CONFIG_TUTORIAL)
        logger = initialize_logging(log_conf_path=log_conf, workflow_conf=self.config)
        return logger

    def _parse_global_vars(self, config, global_vars):
        # substitute global vars throughout the config file, return modified config

        if global_vars is not None:
            config = add_global(config, global_vars, _WE.GLOBAL_VARIABLES)
        config = add_global(
            config,
            get_runtime_global_variables(
                os.path.join(os.getcwd(), "config.json"), os.path.realpath(__file__)
            ),
            _WE.GLOBAL_VARIABLES,
        )
        return config

    def execute(self):
        self._initialize_logging()
        self.workflow.initialize()
        self.workflow.execute()

        sys.exit(0)



from copy import deepcopy
from typing import List

from icolos.utils.enums.step_enums import StepBaseEnum, StepFepPlusEnum
from icolos.utils.enums.program_parameters import FepPlusEnum
from icolos.utils.execute_external.fep_plus import FepPlusExecutor

from pydantic import BaseModel, PrivateAttr
import os
from icolos.core.workflow_steps.step import _LE
import time
from icolos.core.workflow_steps.schrodinger.fep_base import StepFEPBase

from icolos.utils.general.icolos_exceptions import StepFailed

_FE = FepPlusEnum()
_SFE = StepFepPlusEnum()
_SBE = StepBaseEnum


class StepFepPlusExec(StepFEPBase, BaseModel):
    """
    Execute the FEP+ workflow, interfaced with AWS
    """

    class Config:
        underscore_attrs_are_private = True

    _job_id = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=FepPlusExecutor)
        self._check_backend_availability()

        self._job_id = None

    def _parse_arguments(self, tmp_dir: str):
        parameters = deepcopy(self.settings.arguments.parameters)
        arguments = []

        if parameters:
            for key in parameters.keys():
                # treat this specially, needs escaped quotations
                arguments.append(key)
                if key == "-solvent_asl":
                    arguments.append(f"'{parameters[key]}'")
                else:
                    arguments.append(str(parameters[key]))
        # for our AWS config, need to set processors per job =1
        if "-ppj" not in arguments:
            arguments.extend(["-ppj", "1"])
            self._logger.log(
                "Set -ppj 1 for AWS execution, since no override was specified",
                _LE.DEBUG,
            )
        if _SFE.RETRIES not in arguments:
            arguments.extend([_SFE.RETRIES, "3"])
        if len(self.settings.arguments.flags) > 0:
            for flag in self.settings.arguments.flags:
                arguments.append(str(flag))
        # only do this for small-molecule FEP
        if "-protein" in arguments:
            arguments.append(
                os.path.join(
                    tmp_dir, self.data.generic.get_argument_by_extension("maegz")
                )
            )
        else:
            arguments.append(_SFE.FMP_OUTPUT_FILE)

        # remove "-WAIT" if it has been set, as this will interfere with the implementation (and might cause issues
        # due to file system write buffering)
        if _SFE.WAIT_FLAG in arguments:
            self._logger.log(
                "Ignoring -WAIT flag for FEP+ execution (this would interfere with the implementation).",
                _LE.WARNING,
            )
            arguments = [arg for arg in arguments if arg != _SFE.WAIT_FLAG]
        return arguments

    def _unit_test_simulate_output(self, fmp_data, log_data):
        # call this method from the unit instead of the execute method to write out the expected output
        tmp_dir = self._make_tmpdir()
        with open(
            os.path.join(
                tmp_dir,
                f"{self.settings.arguments.parameters[_SFE.JOBNAME_FLAG]}_{_SFE.FMP_OUTPUT_FILE}",
            ),
            "w",
        ) as f:
            f.write(fmp_data)
        with open(
            os.path.join(
                tmp_dir,
                f"{self.settings.arguments.parameters[_SFE.JOBNAME_FLAG]}_{_SFE.LOGFILE}",
            ),
            "w",
        ) as f:
            f.write(fmp_data)
        self._parse_output(tmp_dir)
        self._extract_log_file_data(tmp_dir)
        self._remove_temporary(tmp_dir)

    def _get_job_id(self, result):
        parts = str(result.stdout).split("\n")
        for part in parts:
            if _SFE.JOBID_STRING in part:
                # full_job_id looks something like 549a938d-d2ca-11eb-b9f2-0a6713e9bd3a but only the first part of the
                # hash is needed to access the right job afterwards
                full_job_id = part.split(" ")[1]
                self._job_id = full_job_id.split("-")[0]
                self._logger.log(f"JobId of FEP+ run is {self._job_id}.", _LE.DEBUG)
        if self._job_id is None:
            self._logger.log(
                "Could not obtain JobId after execution - abort.", _LE.ERROR
            )
            raise StepFailed

    def _get_log_file(self) -> List[str]:
        arguments = [
            self._job_id,
            _SFE.FILE_NAME,
            f'{self.settings.arguments.parameters[_SFE.JOBNAME_FLAG]}_{_SFE.LOGFILE}"',
        ]
        logging_result = None
        trials = 0
        while trials < 30000:
            logging_result = self._backend_executor.execute(
                command=_FE.JSC_TAIL_FILE, arguments=arguments, check=False
            )
            if logging_result.returncode != 0:
                time.sleep(30)
                trials += 1
                continue
            else:
                break
        if logging_result is None:
            raise StepFailed("Could not obtain log file from server within time limit.")
        log_lines = str(logging_result.stdout).split("\n")
        return log_lines

    def _get_new_lines(self, old_file) -> List[str]:
        new_lines = self._get_log_file()
        # take the first n lines off the new log file where n is the length of the old log file
        diff = new_lines[len(old_file) - 1 :]
        return diff

    def _wait_for_job_completion(self):
        # get the log file at this state
        log_file = self._get_log_file()
        for line in log_file:
            self._logger_blank.log(line, _LE.INFO)
        # TODO: set maximum (or at least allow to set a maximum)
        while (_SFE.FEP_EXEC_COMPLETE not in log_file) and (
            _SFE.FEP_EXEC_PARTIAL_COMPLETE not in log_file
        ):
            time.sleep(30)
            new_lines = self._get_new_lines(log_file)
            if len(new_lines) > 0:
                for line in new_lines:
                    self._logger_blank.log(line, _LE.INFO)
                    log_file.append(line)

    def _clean_up(self, tmp_dir: str):
        self._remove_temporary(tmp_dir)
        self._job_id = None

    def execute(self):
        # generate the temporary directory and populate it with the required files
        tmp_dir = self._make_tmpdir()
        self.data.generic.write_out_all_files(tmp_dir)

        # check compounds loaded in properly
        if not self.data.compounds:
            self._logger.log(
                f"No compounds were loaded for step {self.step_id}!  If this was intentional you can ignore this warning.",
                _LE.WARNING,
            )

        # obtain the arguments as a list of strings
        arguments = self._parse_arguments(tmp_dir)
        self._logger.log(f"Executing FEP+ calculation in {tmp_dir}.", _LE.INFO)

        # execute fep_plus
        self._apply_token_guard()
        result = self._backend_executor.execute(
            command=_FE.FEP_EXECUTOR, arguments=arguments, location=tmp_dir, check=True
        )

        # get job ID from the job server
        self._get_job_id(result)

        # wait for job completion
        self._wait_for_job_completion()

        # extract the edge information from the log file (rather than the annotated map, as this is easier)
        self._parse_output(tmp_dir)
        self._extract_log_file_data(tmp_dir)
        self._logger.log(f"Completed FEP+ execution.", _LE.INFO)

        # clean-up and reset
        self._clean_up(tmp_dir)



# implement pdbfixer as FOSS alternative to proteinprep
from icolos.utils.enums.step_enums import StepPdbFixerEnum
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.program_parameters import PdbFixerEnum
from icolos.utils.execute_external.execute import Executor
from pydantic import BaseModel
from pdbfixer.pdbfixer import PDBFixer
import os


_SFE = StepPdbFixerEnum()
_FE = PdbFixerEnum()


class StepPdbFixer(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=Executor)

    def _parse_arguments(self):
        default_flags = [
            "--replace-nonstandard",
            "--add-residues",
        ]
        default_params = {
            "--ph": "7.0",
            "--add-atoms": "all",
            "--keep-heterogens": "all",
        }
        arguments = []
        for arg in self.settings.arguments.flags:
            arguments.append(arg)
        for key, value in self.settings.arguments.parameters.items():
            formatted_arg = f"{key}={value}"
            arguments.append(formatted_arg)
        for key in default_flags:
            if key not in self.settings.arguments.flags:
                arguments.append(key)
        for key, value in default_params.items():
            if key not in self.settings.arguments.parameters.keys():
                formatted_arg = f"{key}={value}"
                arguments.append(formatted_arg)
        return arguments

    def execute(self):

        tmp_dir = self._make_tmpdir()

        self.data.generic.write_out_all_files(tmp_dir)
        pdb_files = self.data.generic.get_file_names_by_extension("pdb")

        arguments = self._parse_arguments()

        for file in pdb_files:
            path = os.path.join(tmp_dir, file)
            arguments.extend(["--output", path])
            arguments = [path] + arguments

            self._backend_executor.execute(
                command=_FE.FIXER, arguments=arguments, location=tmp_dir, check=True
            )

            #
        self._parse_output(tmp_dir)

        self._remove_temporary(tmp_dir)



from typing import List
from icolos.core.containers.perturbation_map import Edge
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
import os
from icolos.utils.enums.program_parameters import PMXEnum, PMXAtomMappingEnum
from icolos.core.workflow_steps.step import _LE
import numpy as np

from icolos.utils.execute_external.pmx import PMXExecutor
from icolos.utils.general.parallelization import SubtaskContainer

_PE = PMXEnum()
_PAE = PMXAtomMappingEnum()


class StepPMXligandHybrid(StepPMXBase, BaseModel):
    """Ligand alchemy: hybrid structure/topology."""

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(PMXExecutor)

    def _execute_command(self, jobs: List[str]):
        edge = jobs[0]
        parts = edge.split("_")
        lig1 = parts[0]
        lig2 = parts[1]

        arguments = {
            "-i1": os.path.join(
                self.work_dir,
                _PAE.LIGAND_DIR,
                lig1,
                "MOL.pdb",
            ),
            "-i2": os.path.join(
                self.work_dir,
                _PAE.LIGAND_DIR,
                lig2,
                "MOL.pdb",
            ),
            "-itp1": os.path.join(self.work_dir, _PAE.LIGAND_DIR, lig1, "MOL.itp"),
            "-itp2": os.path.join(self.work_dir, _PAE.LIGAND_DIR, lig2, "MOL.itp"),
        }
        # write output files the hybridStrTop directory for each edge
        output_dir = os.path.join(self.work_dir, edge, _PE.HYBRID_STR_TOP)
        arguments = self._prepare_arguments(args=arguments, output_dir=output_dir)
        self._backend_executor.execute(
            command=_PE.LIGANDHYBRID,
            arguments=arguments,
            check=True,
            location=self.work_dir,
        )

    def _prepare_arguments(self, args, output_dir):
        """
        Prepare the final set of arguments as a list, config overrides defaults
        """
        prepared_args = []
        default_args = {
            "-pairs": f"{output_dir}/pairs1.dat",
            "-oA": f"{output_dir}/mergedA.pdb",
            "-oB": f"{output_dir}/mergedB.pdb",
            "-oitp": f"{output_dir}/merged.itp",
            "-offitp": f"{output_dir}/ffmerged.itp",
            "-log": f"{output_dir}/mapping.log",
        }
        for key, value in args.items():
            default_args[key] = value

        for key, value in self.settings.arguments.parameters.items():
            default_args[key] = value

        for key, value in default_args.items():
            prepared_args.append(key),
            prepared_args.append(value)

        for flag in self.settings.arguments.flags:
            prepared_args.append(flag)
        return prepared_args

    def execute(self):
        assert self.work_dir is not None and os.path.isdir(self.work_dir)

        edges = [e.get_edge_id() for e in self.get_edges()]

        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self._execute_command,
            step_id="ligandHybrid",
            result_checker=self._check_result,
        )

    def _check_result(self, batch: List[List[str]]) -> List[List[bool]]:
        """
        Look in each hybridStrTop dir and check the output pdb files exist for the edges
        """
        output_files = ["ffmerged.itp", "mergedB.pdb", "mergedA.pdb"]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(
                                os.path.join(self.work_dir, job, "hybridStrTop", f)
                            )
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results


help_string = """
pmx ligandHybrid -h
usage: pmx [-h] [-i1 lig1.pdb] [-i2 lig2.pdb] [-itp1 lig1.itp]
           [-itp2 lig2.itp] [-pairs pairs.dat] [-n1 scaffold1.ndx]
           [-n2 scaffold2.ndx] [-oA mergedA.pdb] [-oB mergedB.pdb]
           [-oitp merged.itp] [-offitp ffmerged.itp] [-log hybrid.log]
           [--d 0.05] [--fit] [--split] [--scDUMm 1.0] [--scDUMa 1.0]
           [--scDUMd 1.0] [--deAng]

Provided two structures and topologies, build hybrid structure/topology.

optional arguments:
  -h, --help            show this help message and exit
  -i1 lig1.pdb          Input ligand structure 1. Default is "lig1.pdb"
  -i2 lig2.pdb          Input ligand structure 2. Default is "lig2.pdb"
  -itp1 lig1.itp        Input ligand topology 1. Default is "lig1.itp"
  -itp2 lig2.itp        Input ligand topology 2. Default is "lig2.itp"
  -pairs pairs.dat      Optional input: atom pair mapping. 
  -n1 scaffold1.ndx     Optional input: index of atoms to consider for mol1 
  -n2 scaffold2.ndx     Optional input: index of atoms to consider for mol2 
  -oA mergedA.pdb       Output: hybrid structure based on the ligand 1. Default is "mergedA.pdb"
  -oB mergedB.pdb       Output: hybrid structure based on the ligand 2. Default is "mergedB.pdb"
  -oitp merged.itp      Output: hybrid topology. Default is "merged.itp"
  -offitp ffmerged.itp  Output: atomtypes for hybrid topology. Default is "ffmerged.itp"
  -log hybrid.log       Output: log file. Default is "hybrid.log"
  --d 0.05              Optional: if -pairs not provided, distance (nm) between atoms to consider them morphable for alignment approach (default 0.05 nm).
  --fit                 Fit mol2 onto mol1, only works if pairs.dat is provided
  --split               split the topology into separate transitions
  --scDUMm 1.0          scale dummy masses using the counterpart atoms
  --scDUMa 1.0          scale bonded dummy angle parameters
  --scDUMd 1.0          scale bonded dummy dihedral parameters
  --deAng               decouple angles composed of 1 dummy and 2 non-dummies
"""



from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel


class StepPMXgenlib(StepPMXBase, BaseModel):
    """Generate pmx ff library."""

    def __init__(self, **data):
        super().__init__(**data)

    def execute(self):
        pass


help_string = """
pmx genlib -h
usage: pmx [-h] [-f1 ipdb1] [-f2 ipdb2] [-o1 opdb1] [-o2 opdb2]
           [--ffpath ffpath] [--fatp fatp] [--fnb fnb] [--moltype moltype]
           [--noalign] [--cbeta] [--noH2Heavy] [--log log]

The script creates hybrid structure and topology database entries (mtp and rtp)
in order to generate a pmx alchemical force field library.

The easiest way to generate the library is to call this script from within
the folder of the force field you are interested in.

If two pdb files (aligned on the backbone) are provided, the hybrid pdb, mtp,
and rtp files are written to file. If no pdb input file is provided,
the script uses pregenerated residues in order to build hybrid pdb, mtp, and
rtp files for all possible residue pairs, thus preparing the whole pmx ff
library.

In addition, atomtype (-fatp) and non-bonded parameter (-fnm) files for the
introduced dummy atoms are generated. By default, these point towards the
files already present in the forcefield. In this way, the additional parameters
for the dummies are appended to the existing ff file, rather than being
written to new files.

optional arguments:
  -h, --help         show this help message and exit
  -f1 ipdb1          First input PDB file. Default is none provided.
  -f2 ipdb2          Second input PDB file. Default is none provided.
  -o1 opdb1          First output PDB file. Default is none provided.
  -o2 opdb2          Second output PDB file. Default is none provided.
  --ffpath ffpath    Path to mutation forcefield. Default is current folder.
  --fatp fatp        Atom types (atp) file. If the file is 
                     present, data is appended to it, otherwise a new 
                     file is created. Default is "atomtypes.atp".
  --fnb fnb          Non-bonded (nb) types file. If the file is 
                     present, data is appended to it, otherwise a new 
                     file is created. Default is "ffnonbonded.itp".
  --moltype moltype  The type of molecule for which the library is 
                     being built. Available options are "protein", "dna", 
                     or "rna". Default is "protein".
  --noalign          Whether to align the sidechains of the two 
                     input PDB files provided. Default it True; this flag 
                     sets it to False.
  --cbeta            Whether to morph sidechain between the two 
                     residues or to use dummy atoms to (de)couple the 
                     whole sidechain. By default, sidechain atoms are 
                     morphed so to minimise the size of the perturbation. 
                     With this flag set, whole sidechains are (de)coupled 
                     instead; i.e. all atoms after C-beta are not mapped 
                     between the two residues.
  --noH2Heavy        Whether to allow hydrogen to/from heavy atoms 
                     morphing. Default is True, this flag sets it to False.
  --log log          Logging level. Either "info" or "debug". Default is "info".
"""



import logging

from icolos.loggers.base_logger import BaseLogger


class AgentLogger(BaseLogger):
    def __init__(self):
        super().__init__()

    def _initialize_logger(self):
        logger = logging.getLogger(self._LE.LOGGER_AGENT)
        return logger



import os
import json
import pickle
from icolos.core.workflow_steps.active_learning.surrogate_model import SurrogateModel
from icolos.core.workflow_steps.active_learning.acquisition_function import (
    AcquisitionFunction,
)
from icolos.core.workflow_steps.step import _LE
from icolos.utils.enums.step_enums import StepBaseEnum, StepGlideEnum
from icolos.utils.enums.composite_agents_enums import WorkflowEnum
from icolos.utils.enums.step_enums import (
    StepActiveLearningEnum,
)
from icolos.core.workflow_steps.step import StepBase
from icolos.core.composite_agents.workflow import WorkFlow

from rdkit import Chem
from rdkit.Chem import AllChem
import pandas as pd

_SALE = StepActiveLearningEnum()
_WE = WorkflowEnum()
_SBE = StepBaseEnum
_SGE = StepGlideEnum()


class StepProspectiveREINVENT(StepBase):
    """
    Class to run prospective active learning with REINVENT.
    The scoring function component (e.g., docking via LigPrep + Glide) will be predicted by a surrogate model
    for a fraction of every batch of SMILES generated by REINVENT at a given epoch. The goal is to mitigate
    computational costs of expensive components. The corresponding fraction of SMILES not predicted will be
    sent to the defined oracle and the ground truth labels will be used to augment the surrogate model.
    This process is repeated to constitute an active learning loop
    """

    def __init__(self, **data):
        super().__init__(**data)

    def _run(self, original_smiles: list, save_path: str):
        """
        main method controlling what icolos logic to run out of 3 possibilities:
        1. run normal REINVENT (corresponding to warm-up)
        2. run normal REINVENT with pooling (corresponding to initial pooling epochs)
        3. run active learning REINVENT
        """

        (
            epochs,
            warmup,
            retrain,
            initial_pooling_epochs,
            acquisition_batch_size,
            acquisition_function_name,
            surrogate_model_type,
            oracle_config,
        ) = self._get_run_parameters()

        # extract current epoch
        curr_epoch = self._get_curr_epoch()

        if curr_epoch <= warmup:
            # run "normal" REINVENT, i.e., all generated SMILES are passed to the Oracle
            self._run_normal_reinvent(
                curr_epoch=curr_epoch,
                save_path=save_path,
                original_smiles=original_smiles,
                oracle_config=oracle_config
            )

        elif curr_epoch <= (warmup + initial_pooling_epochs):
            # continue running "normal" REINVENT but start pooling the (SMILES, docking score) pairs
            # to generate the initial training set for the surrogate model
            self._run_normal_reinvent_with_pooling(
                curr_epoch=curr_epoch,
                save_path=save_path,
                original_smiles=original_smiles,
                oracle_config=oracle_config,
            )

        # run active learning REINVENT for the remainder of the experiment
        else:
            self._run_active_learning_reinvent(
                curr_epoch=curr_epoch,
                save_path=save_path,
                original_smiles=original_smiles,
                oracle_config=oracle_config,
                warmup=warmup,
                initial_pooling_epochs=initial_pooling_epochs,
                retrain=retrain,
                surrogate_model_type=surrogate_model_type,
                acquisition_function_name=acquisition_function_name,
                acquisition_batch_size=acquisition_batch_size,
            )

    def _get_morgan_fingerprints(self, smiles) -> list:
        """returns Morgan fingerprints for a list of smiles"""
        molecules = [Chem.MolFromSmiles(s) for s in smiles]
        return [
            Chem.AllChem.GetMorganFingerprintAsBitVect(mol, radius=3, nBits=2048)
            for mol in molecules
        ]

    def _get_curr_epoch(self) -> int:
        """get the current REINVENT epoch number to determine which stage of active learning to execute"""

        global_variables = self.get_workflow_object().header.global_variables
        curr_epoch = int(global_variables["step_id"])
        # curr_epoch + 1 to start at epoch 1
        return curr_epoch + 1

    def _save_state(
        self,
        curr_epoch: int,
        save_path: str,
        original_smiles: list,
        fingerprints: list,
        labels: list,
        pooled_smiles=None,
        pooled_labels=None,
        pooled_fingerprints=None,
        surrogate=None,
        save_pool=False,
        non_acquired_smiles=None,
        non_acquired_labels=None,
    ):
        """stores intermediate results after every epoch"""
        # save the pooled compounds to be used as training data for the surrogate model
        if save_pool:
            # make sure pooled_compounds are smiles, but we should also save the enumerated forms too?
            pooled_smiles.extend(original_smiles)
            pooled_labels.extend(labels)
            df = pd.DataFrame({"smiles": pooled_smiles, "labels": pooled_labels})
            df.to_csv(os.path.join(save_path, f"pooled_data/epoch_{curr_epoch}.csv"))

            # if using PI or EI, save the current best
            af = self._get_additional_setting(_SALE.ACQUISITION_FUNCTION, default=_SALE.RANDOM)
            if (af == _SALE.PI) or (af == _SALE.EI):
                current_best = min(pooled_labels)
                with open("current_best.txt", "w+") as f:
                    f.write(f"{str(current_best)}\n{str(curr_epoch)}")

            pooled_fingerprints.extend(fingerprints)
            pkl_path = os.path.join(save_path, f"pooled_data/epoch_{curr_epoch}_fingerprints.sav")
            pickle.dump(pooled_fingerprints, open(pkl_path, "wb"))

        # save the non-acquired SMILES and the predicted labels, if applicable
        if non_acquired_smiles is not None:
            df = pd.DataFrame(
                {"smiles": non_acquired_smiles, "labels": non_acquired_labels}
            )
            df.to_csv(
                os.path.join(save_path, f"predicted_smiles/epoch_{curr_epoch}.csv")
            )

        # save the surrogate model state
        if isinstance(surrogate, SurrogateModel):
            pkl_path = os.path.join(
                save_path, f"surrogate_states/epoch_{curr_epoch}.sav"
            )
            pickle.dump(surrogate, open(pkl_path, "wb"))

    def _read_state(self, curr_epoch, save_path: str, train_negative_pool: False):
        """extract the previous epoch's intermediate results to be used in the current epoch"""
        # read the saved state CSV file and extract the pooled SMILES and labels
        # at the first pooling epoch, there will be neither pooled compounds nor surrogate state
        try:
            # curr_epoch - 1 to ensure the previous state prior to the current active learning iteration is read
            df = pd.read_csv(
                os.path.join(save_path, f"pooled_data/epoch_{curr_epoch-1}.csv")
            )
            pooled_smiles, pooled_labels = list(df["smiles"]), list(df["labels"])

            # ------------------------------------------------------------------------------------
            # TODO: temporary implementation to train on warm-up data
            if train_negative_pool:
                files = os.listdir("docking_scores")
                files = sorted(files, key=lambda x: int(x.split("_")[0]))
                for file in files:
                    df = pd.read_csv(os.path.join("docking_scores", file))
                    pooled_smiles.extend(df["original_smiles"])
                    pooled_labels.extend(df["docking_score"])
                self._logger.log("added oracle scores", _LE.DEBUG)
                pooled_fingerprints = self._get_morgan_fingerprints(smiles=pooled_smiles)
            # --------------------------------------------------------------------------------------
            else:
                pkl_path = os.path.join(save_path, f"pooled_data/epoch_{curr_epoch-1}_fingerprints.sav")
                pooled_fingerprints = pickle.load(open(pkl_path, 'rb'))
                # TODO: could replace these magic strings

        except Exception:
            pooled_smiles, pooled_fingerprints, pooled_labels = [], [], []

        try:
            surrogate = pickle.load(
                open(
                    os.path.join(
                        save_path, f"surrogate_states/epoch_{curr_epoch-1}.sav"
                    ),
                    "rb",
                )
            )
        except Exception:
            surrogate = None

        return pooled_smiles, pooled_fingerprints, pooled_labels, surrogate

    def _query_oracle(self, save_path: str, original_smiles, oracle_config) -> list:
        """executes the oracle Workflow object and returns the scores"""

        oracle_workflow = self._construct_oracle_workflow(oracle_config=oracle_config)
        # copy and save the REINVENT's batch of SMILES to a temporary save folder
        self._create_manipulable_compounds_file(
            original_smiles=original_smiles, save_path=save_path
        )
        oracle_workflow.execute()

        # extract the best scores per compound and its enumerations
        scores = self._extract_scores_from_oracle(
            original_smiles=original_smiles, oracle_workflow=oracle_workflow
        )

        return scores

    def _create_manipulable_compounds_file(self, original_smiles, save_path: str):
        """
        this method copies the compounds generated by REINVENT into a new directory.
        The compounds here can be freely manipulated, in particular, partitioning into
        compounds to acquire and compounds to predict via the surrogate model
        """

        path = os.path.join(save_path, "reinvent_batches")
        with open(os.path.join(path, "oracle_compounds.smi"), "w+") as f:
            for smiles in original_smiles:
                f.write(f"{smiles}\n")

    def _extract_scores_from_oracle(
        self, original_smiles, oracle_workflow: WorkFlow
    ) -> list:
        """
        extracts the "best" score from the oracle and is particularly relevant in cases where more than 1 score
        is given to a SMILES. This is usually taken care of by icolos write-out and is done here only because
        the training data for the surrogate model (SMILES, score pairs) needs to be stored
        """
        scores = []
        for smiles, compound in zip(
            original_smiles, oracle_workflow.get_steps()[-1].get_compounds()
        ):
            # TODO: assumes lower the score the better. Add a parameter in the JSON to control this for generalizability
            try:
                    scores_tracker = []
                    for enumeration in compound.get_enumerations():
                        for conformer in enumeration.get_conformers():
                            # docking scores here are with Epik corrections
                            # TODO: docking score is hard-coded at the moment, need an icolos parameter to specifiy this
                            score = float(
                                conformer.get_molecule().GetProp(_SGE.GLIDE_DOCKING_SCORE)
                            )
                            scores_tracker.append(score)
                    # TODO: this may cause an error if list is empty --> enumeration not added in dummy conf?
                    #  try block handles this for the time being
                    scores.append(min(scores_tracker))
            except Exception:
                scores.append(float(0.0))

        return scores

    def _write_reinvent_feedback(self, original_smiles, labels):
        """this method manually writes the REINVENT feedback"""
        path = self.get_workflow_object().header.global_variables["output_json_path"]

        # below is the form expected by REINVENT
        # extract the oracle label
        values_key = self._get_additional_setting(_SALE.ORACLE_LABEL)
        try:
            output_dict = {
                "results": [{"values_key": values_key, "values": labels}],
                "names": [idx for idx in range(len(original_smiles))],
            }
        # if for some reason, the number of SMILES is not equal to the number of labels, labels (0.0 float)
        # are manually appended to prevent code error. This is a temporary fix for the initial experiments
        # and one should ensure in the icolos step that each SMILES is accounted for
        except Exception:
            missing_values = len(original_smiles) - len(values_key)
            append_values = ["dummy SMILES" for idx in range(missing_values)]
            values_key.extend(append_values)

            missing_labels = len(original_smiles) - len(labels)
            append_labels = [0.0 for idx in range(missing_labels)]
            labels.extend(append_labels)

            output_dict = {
                "results": [{"values_key": values_key, "values": labels}],
                "names": [idx for idx in range(len(original_smiles))],
            }

            self._logger.log("REINVENT Feedback Missing Values. Manually Added.", _LE.DEBUG)

        with open(path, "w+") as f:
            json.dump(output_dict, f, indent=2)

    def _run_normal_reinvent(
        self, curr_epoch, save_path, original_smiles, oracle_config
    ):
        """runs a normal REINVENT epoch"""
        scores = self._query_oracle(
            save_path=save_path,
            original_smiles=original_smiles,
            oracle_config=oracle_config,
        )

        self._save_state(
            curr_epoch=curr_epoch,
            original_smiles=original_smiles,
            fingerprints=None,
            labels=scores,
            save_path=save_path,
        )
        self._write_reinvent_feedback(original_smiles=original_smiles, labels=scores)

    def _run_normal_reinvent_with_pooling(
        self, curr_epoch, save_path: str, original_smiles, oracle_config
    ):
        """runs a normal REINVENT epoch and stores all (SMILES, scores) pairs as training data for the surrogate"""
        # read the current state
        pooled_smiles, pooled_fingerprints, pooled_labels, surrogate = self._read_state(
            curr_epoch=curr_epoch, save_path=save_path, train_negative_pool=False
        )

        scores = self._query_oracle(
            save_path=save_path,
            original_smiles=original_smiles,
            oracle_config=oracle_config,
        )

        # generate the Morgan fingerprints for saving
        fingerprints = self._get_morgan_fingerprints(smiles=original_smiles)

        self._save_state(
            curr_epoch=curr_epoch,
            original_smiles=original_smiles,
            fingerprints=fingerprints,
            labels=scores,
            pooled_smiles=pooled_smiles,
            pooled_labels=pooled_labels,
            pooled_fingerprints=pooled_fingerprints,
            save_pool=True,
            save_path=save_path,
        )

        self._write_reinvent_feedback(original_smiles=original_smiles, labels=scores)

    def _run_active_learning_reinvent(
        self,
        curr_epoch,
        save_path: str,
        original_smiles,
        oracle_config,
        warmup,
        initial_pooling_epochs,
        retrain,
        surrogate_model_type,
        acquisition_function_name,
        acquisition_batch_size,
    ):
        """runs active learning REINVENT"""

        # the if-else block below was a quick implementation of training the initial surrogate model state
        # using the "negative" data during warm-up. The train_negative_pool boolean of _read_state controls this.
        # In the current code state, it is set to False in both the if-else blocks but could be set to True in
        # the if block to train with negative data
        # TODO: should have a boolean parameter in the icolos json specify whether or not to train with negative data
        if curr_epoch == (warmup + initial_pooling_epochs + 1):
            pooled_smiles, pooled_fingerprints, pooled_labels, surrogate = self._read_state(
                curr_epoch=curr_epoch, save_path=save_path, train_negative_pool=False
            )
        else:
            pooled_smiles, pooled_fingerprints, pooled_labels, surrogate = self._read_state(
                curr_epoch=curr_epoch, save_path=save_path, train_negative_pool=False
            )

        # model training is controlled by the "retrain" parameter. The only exception to this is when
        # curr_epoch = (warmup + initial_pooling_epochs + 1), i.e., the epoch right after initial pooling
        # this epoch corresponds to the initial training of the surrogate model and is also the first epoch
        # where we begin saving the pickled surrogate model
        if (
            curr_epoch == (warmup + initial_pooling_epochs + 1)
            or curr_epoch % retrain == 0
        ):
            surrogate = SurrogateModel(model_type=surrogate_model_type)
            surrogate.fit(pooled_fingerprints, pooled_labels)
        # at this point, whether the surrogate was retrained or not, partition the compounds into acquired and
        # non-acquired. Send acquired to oracle, predict the rest, and send the concatenated data back to REINVENT
        acquisition_function = AcquisitionFunction(
            surrogate=surrogate,
            function=acquisition_function_name,
            acquisition_batch_size=acquisition_batch_size,
        )

        # select points to acquire based on acquisition function
        acquired_smiles, non_acquired_smiles = acquisition_function.partition_compounds(
            original_smiles=original_smiles, fingerprints_pool=pooled_fingerprints
        )

        # send the acquired compounds to the oracle
        acquired_labels = self._query_oracle(
            save_path=save_path,
            original_smiles=acquired_smiles,
            oracle_config=oracle_config,
        )
        # predict the non-acquired compounds and cast to type list from np.array for compatibility with other methods
        non_acquired_labels = list(
            surrogate.predict(self._get_morgan_fingerprints(smiles=non_acquired_smiles))
        )

        # generate the Morgan fingerprints of the acquired compounds for saving
        acquired_fingerprints = self._get_morgan_fingerprints(smiles=acquired_smiles)

        # save the current state for the next iteration
        # note: original_smiles and labels here correspond to the ***acquired*** data.
        # Concatenating the acquired data to the current pool is handled by save_state.
        # The non-acquired data is also saved for analysis at the end of the experiment
        self._save_state(
            curr_epoch=curr_epoch,
            save_path=save_path,
            original_smiles=acquired_smiles,
            fingerprints=acquired_fingerprints,
            labels=acquired_labels,
            pooled_smiles=pooled_smiles,
            pooled_labels=pooled_labels,
            pooled_fingerprints=pooled_fingerprints,
            non_acquired_smiles=non_acquired_smiles,
            non_acquired_labels=non_acquired_labels,
            surrogate=surrogate,
            save_pool=True,
        )

        # concatenate the acquired and non-acquired points to provide REINVENT feedback
        # from REINVENT's perspective, it has no notion of acquired vs. predicted. It simply receives scores back
        concatenated_smiles = acquired_smiles + non_acquired_smiles
        concatenated_labels = acquired_labels + non_acquired_labels
        self._write_reinvent_feedback(
            original_smiles=concatenated_smiles, labels=concatenated_labels
        )

    def _get_run_parameters(self):
        """returns all the experiment parameters"""
        # default 1000 REINVENT epochs if running standard reinforcement learning for small molecules
        epochs = self._get_additional_setting(_SALE.EPOCHS, default=1000)
        warmup = self._get_additional_setting(_SALE.WARMUP, default=200)
        retrain = self._get_additional_setting(_SALE.RETRAIN, default=5)
        initial_pooling_epochs = self._get_additional_setting(
            _SALE.INITIAL_POOLING_EPOCHS, default=5
        )
        acquisition_batch_size = self._get_additional_setting(
            _SALE.ACQUISITION_BATCH_SIZE, default=40
        )
        acquisition_function = self._get_additional_setting(
            _SALE.ACQUISITION_FUNCTION, default=_SALE.RANDOM
        )
        surrogate_model_type = self._get_additional_setting(
            _SALE.SURROGATE_MODEL_TYPE, default=_SALE.RANDOM_FOREST_REGRESSOR
        )
        oracle_config = self._get_additional_setting(_SALE.ORACLE_CONFIG)

        return (
            epochs,
            warmup,
            retrain,
            initial_pooling_epochs,
            acquisition_batch_size,
            acquisition_function,
            surrogate_model_type,
            oracle_config,
        )

    def _construct_oracle_workflow(self, oracle_config) -> WorkFlow:
        """constructs an icolos Workflow object for the oracle (essentially a sub-Workflow)"""

        # load oracle configuration
        with open(oracle_config, "r") as file:
            conf = file.read().replace("\r", "").replace("\n", "")
            conf = json.loads(conf)

        # generate oracle workflow object
        oracle_workflow = WorkFlow(**conf[_WE.WORKFLOW])
        # copy over the header from the parent workflow to extract the icolos global variables
        header = self.get_workflow_object().header
        oracle_workflow.header = header
        oracle_workflow.initialize()

        return oracle_workflow

    def _create_save_folders(self, save_path: str):
        """creates necessary directories for storage of intermediate results"""
        if not os.path.exists(save_path):
            os.mkdir(save_path)

        if not os.path.exists(os.path.join(save_path, "pooled_data")):
            os.mkdir(os.path.join(save_path, "pooled_data"))

        if not os.path.exists(os.path.join(save_path, "predicted_smiles")):
            os.mkdir(os.path.join(save_path, "predicted_smiles"))

        if not os.path.exists(os.path.join(save_path, "surrogate_states")):
            os.mkdir(os.path.join(save_path, "surrogate_states"))

        if not os.path.exists(os.path.join(save_path, "reinvent_batches")):
            os.mkdir(os.path.join(save_path, "reinvent_batches"))

    def execute(self):
        # extract the original SMILES from the REINVENT batch
        original_smiles = [
            compound.get_enumerations()[0].get_original_smile()
            for compound in self.data.compounds
        ]
        # extract the save directory specified to store intermediate results
        save_path = self._get_additional_setting(_SALE.SAVE_DIR)
        # create save folders to store pooled SMILES, pooled fingerprints, and surrogate model states
        self._create_save_folders(save_path=save_path)

        self._run(original_smiles=original_smiles, save_path=save_path)



def get_progress_bar_string(
    done, total, prefix="", suffix="", decimals=1, length=100, fill="█"
):
    percent = ("{0:." + str(decimals) + "f}").format(100 * (done / float(total)))
    filledLength = int(length * done // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    return f"{prefix}|{bar}| {percent}% {suffix}"



from rdkit import Chem


def to_smiles(mol, isomericSmiles=False):
    """
    Converts a Mol object into a canonical SMILES string.
    :param mol: Mol object.
    :return: A SMILES string.
    """
    return Chem.MolToSmiles(mol, isomericSmiles=isomericSmiles)



from icolos.utils.enums.program_parameters import OpenBabelEnum
from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from icolos.loggers.steplogger import StepLogger
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.utils.general.icolos_exceptions import StepFailed
from icolos.utils.enums.logging_enums import LoggingConfigEnum

_LE = LoggingConfigEnum()
_OE = OpenBabelEnum()


class OBabelStructConvert:
    """
    Utility for converting structure files with Schrodinger's StructConvert
    """

    def __init__(
        self,
    ) -> None:
        self._logger = StepLogger()

        self.executor = OpenBabelExecutor()
        if not self.executor.is_available():
            # this shouldn't fail, obabel is in the env
            raise StepFailed("Cannot initialize sdconvert backend - abort.")
        self._logger.log(f"Checked obabel availability - valid.", _LE.DEBUG)

    def execute(self, arguments: list):
        execution_result = self.executor.execute(
            command=_OE.OBABEL, arguments=arguments, check=True
        )
        if execution_result.returncode != 0:
            self._logger.log(
                f"Could not execute sdconvert (returncode != 0) with error: {execution_result.stderr}.",
                _LE.ERROR,
            )

    def convert(self, input_file: str, output_file: str):
        arguments = [
            input_file,
            output_file,
        ]
        self.execute(arguments=arguments)

    def sdf2pdb(self, sdf_file: str, pdb_file: str):
        assert sdf_file.endswith(".sdf")
        assert pdb_file.endswith(".pdb")

        arguments = ["-isdf", sdf_file, "-O", pdb_file]
        self.execute(arguments=arguments)



import os
from typing import List

from pydantic import BaseModel
from rdkit import Chem
from copy import deepcopy
from icolos.utils.execute_external.omega import OMEGAExecutor
from icolos.core.workflow_steps.step import _LE, _CTE
from icolos.utils.general.molecules import get_charge_for_molecule

from icolos.core.containers.compound import Enumeration, Conformer

from icolos.utils.enums.program_parameters import OMEGAEnum, OMEGAOutputEnum
from icolos.core.workflow_steps.confgen.base import StepConfgenBase

_EE = OMEGAEnum()
_COE = OMEGAOutputEnum()


class StepOmega(StepConfgenBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=OMEGAExecutor)
        self._check_backend_availability()

    def _parse_OMEGA_result(
        self, dir_path: str, enumeration: Enumeration
    ) -> List[Conformer]:
        # OMEGA will output a variety of files to "dir_path"
        conformers_sdf = os.path.join(dir_path, _COE.OUTPUT_SDF_NAME)

        # energies are added as a tag in the output
        mol_supplier = Chem.SDMolSupplier(conformers_sdf, removeHs=False)
        charge = str(
            get_charge_for_molecule(enumeration.get_molecule(), add_as_tag=False)
        )
        result = []
        for mol_id, mol in enumerate(mol_supplier):
            mol.SetProp(
                _CTE.CONFORMER_ENERGY_TAG, mol.GetProp(_COE.CLASSIC_ENERGY_OUTPUT_TAG)
            )
            mol.ClearProp(_COE.CLASSIC_ENERGY_OUTPUT_TAG)
            mol.SetProp(_CTE.FORMAL_CHARGE_TAG, charge)
            conf = Conformer(conformer=mol)
            result.append(conf)
        return result

    def _set_input_output_paths(self, parameters: dict, input_path: str) -> dict:
        # this is handled this way to overwrite any specifications from the user for the input / output paths as well
        parameters[_EE.CLASSIC_INPUT] = input_path
        parameters[_EE.CLASSIC_OUTPUT] = _COE.OUTPUT_SDF_NAME
        return parameters

    def _prepare_settings(self, tmp_dir: str, enumeration: Enumeration) -> list:
        # the first argument is the mode of binary "oeomega" (for now defaults to "classic")
        settings = [_EE.OMEGA_MODE_CLASSIC]

        # add flags
        # make sure, the energy tag is set as well
        for flag in self.settings.arguments.flags:
            settings.append(flag)
        if _EE.CLASSIC_SDENERGY not in settings:
            settings.append(_EE.CLASSIC_SDENERGY)

        # add parameters
        parameters = deepcopy(self.settings.arguments.parameters)

        # update / over-write fields that need a specific value or are defined elsewhere
        parameters = self._set_input_output_paths(
            parameters=parameters,
            input_path=self._prepare_temp_input(tmp_dir, enumeration.get_molecule()),
        )

        # flatten the dictionary into a list for command-line execution
        for key in parameters.keys():
            settings.append(key)
            settings.append(parameters[key])
        return settings

    def execute(self):
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if not self._input_object_valid(enumeration):
                    continue

                # set up
                tmp_dir = self._move_to_temp_dir()
                settings = self._prepare_settings(tmp_dir, enumeration=enumeration)

                # execution
                self._logger.log(
                    f"Executing OMEGA backend in folder {tmp_dir}.", _LE.DEBUG
                )
                self._backend_executor.execute(
                    command=_EE.OMEGA, arguments=settings, check=False
                )
                self._restore_working_dir()

                # parsing
                conformers = self._parse_OMEGA_result(tmp_dir, enumeration=enumeration)
                enumeration.clear_conformers()
                enumeration.add_conformers(conformers=conformers, auto_update=True)
                self._logger.log(
                    f"Completed OMEGA for enumeration {enumeration.get_index_string()}, added {len(conformers)} conformers.",
                    _LE.INFO,
                )

                # clean-up
                self._remove_temporary(tmp_dir)



import shutil
from icolos.core.containers.gmx_state import GromacsState
from icolos.utils.enums.program_parameters import (
    GromacsEnum,
)
from icolos.utils.enums.step_enums import StepGromacsEnum, StepOpenFFEnum
from pydantic import BaseModel
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.execute_external.execute import Executor
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.core.workflow_steps.step import _LE
import os
import re
from string import ascii_uppercase
from rdkit import Chem
from openmm.app import PDBFile
from parmed.gromacs import GromacsTopologyFile
from openff.toolkit.topology import Molecule, Topology
from openff.toolkit.typing.engines.smirnoff import ForceField
import parmed as pmd

_SGE = StepGromacsEnum()
_GE = GromacsEnum()
_SOFE = StepOpenFFEnum()


class StepGMXPdb2gmx(StepGromacsBase, BaseModel):
    class Config:
        arbitrary_types_allowed = True

    _shell_executor: Executor = None
    _antechamber_executor: Executor = None
    _acpype_executor: Executor = None
    _schrodinger_executor: SchrodingerExecutor = None

    def __init__(self, **data):
        """
        Executes system parametrisation for gromacs MD setup
        Generates GAFF params for unknown components with Antechamber
        """
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()
        self._shell_executor = Executor()
        self._antechamber_executor = Executor()

    def _split_protein_ligand_complex(self, tmp_dir):
        # split the file into protein and an arbitrary number of ligands and cofactors
        # Handle for multiple cofactors of the same type
        struct_file = [
            file for file in os.listdir(tmp_dir) if file.endswith(_SGE.FIELD_KEY_PDB)
        ][0]
        with open(os.path.join(tmp_dir, struct_file), "r") as f:
            data = f.readlines()
        # handles arbitrary number of ligands, cofactors, etc
        ligand_lines = {}
        protein_lines = []

        for line in data:
            parts = line.upper().split()

            # filter header lines etc
            if len(parts) > 4 and parts[0] in _GE.ATOMS:

                # catch the easy cases where there is a direct match to the parametrised components against internal dict
                if (
                    parts[3] in _GE.AMBER_PARAMETRISED_COMPONENTS
                    or parts[3] in _GE.IONS
                ):

                    protein_lines.append(line)

                # catch cases where ions have non-standard residue names e.g. NA3
                elif parts[3][:2] in _GE.IONS and re.findall(
                    re.compile(rf"{parts[3][:2]}[0-9]+"), line
                ):

                    pattern = rf"{parts[3][:2]}[0-9]+"
                    pattern = re.compile(pattern)

                    line = re.sub(pattern, parts[3][:2], line)
                    protein_lines.append(line)

                else:
                    # component is not parametrised, add to the ligands
                    if parts[4] in list(ascii_uppercase):
                        try:
                            ligand_lines[f"{parts[3]}:{parts[5]}"].append(line)
                        except KeyError:
                            # ligand key not created yet, identify by chain + res num to handle multiple identical components
                            ligand_lines[f"{parts[3]}:{parts[5]}"] = [line]
                    else:  # the 5th col index is the first coord col
                        try:
                            ligand_lines[f"{parts[3]}:{parts[4]}"].append(line)
                        except KeyError:
                            ligand_lines[f"{parts[3]}:{parts[4]}"] = [line]

        for key, value in ligand_lines.items():
            # write ligand components as separate pdb files
            with open(os.path.join(tmp_dir, f"{key}.pdb"), "w") as f:
                f.writelines(value)
        with open(os.path.join(tmp_dir, _SGE.PROTEIN_PDB), "w") as f:
            f.writelines(protein_lines)
        self._remove_temporary(os.path.join(tmp_dir, struct_file))
        return list(ligand_lines.keys())

    def _generate_gaff_params(
        self, tmp_dir: str, input_pdb: str, topol: GromacsState
    ) -> None:
        """
        :param tmp_dir: step's base directory
        :param input_pdb: file name for the ligand being parametrised

        Produces a gmx ITP file for the ligand, and updates topology
        """
        # main pipeline for producing GAFF parameters for a ligand
        charge_method = self._get_additional_setting(
            key=_SGE.CHARGE_METHOD, default="bcc"
        )
        conf = Chem.rdmolfiles.MolFromPDBFile(os.path.join(tmp_dir, input_pdb))
        formal_charge = Chem.rdmolops.GetFormalCharge(conf) if conf is not None else 0
        self._logger.log(
            f"Computed formal charge: {formal_charge} for structure {input_pdb}",
            _LE.DEBUG,
        )
        ligand_stub = input_pdb.split(".")[0]
        # Step 4: run the acpype script to generate the ligand topology file for GAFF
        self._logger.log(
            f"Running acpype on structure:{input_pdb}, charge method: {charge_method}",
            _LE.DEBUG,
        )
        acpype_args = [
            "-di",
            input_pdb,
            "-c",
            charge_method,
            "-o",
            "gmx",
            "-n",
            formal_charge,
        ]
        self._antechamber_executor.execute(
            command=_GE.ACPYPE_BINARY,
            arguments=acpype_args,
            location=tmp_dir,
            check=True,
        )

        acpype_dir = os.path.join(tmp_dir, f"{ligand_stub}.acpype")
        # TODO: refactor this
        lig_itp = [
            f
            for f in os.listdir(os.path.join(tmp_dir, acpype_dir))
            if f.endswith("GMX.itp")
        ][0]
        lig_pdb = [
            f
            for f in os.listdir(os.path.join(tmp_dir, acpype_dir))
            if f.endswith("pdb")
        ][0]

        topol.add_itp(os.path.join(tmp_dir, acpype_dir), [lig_itp], gen_posre=True)
        topol.add_molecule(lig_itp.split("_")[0], 1)
        topol.append_structure(
            os.path.join(tmp_dir, acpype_dir), lig_pdb, sanitize=True
        )

    def _generate_openff_params(
        self, tmp_dir: str, input_pdb: str, topol: GromacsState
    ):
        """
        Generate Amber-compatible Smirnoff params for a single component

        """
        # TODO: ensure we can handle parameter generation for multiple unique mol types
        # get the mols
        mols = [Molecule.from_smiles(self._get_additional_setting(_SOFE.UNIQUE_MOLS))]
        pdb_file = PDBFile(os.path.join(self._get_additional_setting("lig_pdb")))
        omm_topology = pdb_file.topology

        off_topology = Topology.from_openmm(omm_topology, unique_molecules=mols)

        forcefield = ForceField(self._get_additional_setting(_SOFE.FORCEFIELD))

        omm_system = forcefield.create_openmm_system(off_topology)

        parmed_struct = pmd.openmm.load_topology(
            omm_topology, system=omm_system, xyz=pdb_file.positions
        )
        parmed_struct.save(os.path.join(tmp_dir, "MOL.top"), overwrite=True)
        parmed_struct.save(os.path.join(tmp_dir, "MOL.pdb"), overwrite=True)
        gmx_top = GromacsTopologyFile().from_structure(parmed_struct)
        gmx_top.write(os.path.join(tmp_dir, "MOL.itp"), itp=True)
        topol.add_itp(path=tmp_dir, files=["MOL.itp"])
        topol.generate_posre(path=tmp_dir, itp_file="MOL.itp")
        topol.append_structure(path=tmp_dir, file="MOL.pdb", sanitize=True)

    def execute(self):
        """Takes in a ligand pdb file and generates the required topology, based on the backend specified in the config json.
        Currently supported AnteChamber

        Execution looks like this currently:
        (1) split the protein from the other components
        (2) generate the topology for the protein separately
        (4) run the parametrisation pipeline on each HET component in serial, generating either GAFF or SMIRNOFF params
        (5) combine the topologies

        """

        tmp_dir = self._make_tmpdir()
        topol = self.get_topol()
        self.write_input_files(tmp_dir)
        lig_ids = self._split_protein_ligand_complex(tmp_dir)
        self._logger.log(
            f"Parameters will be generated for the following components: {str(lig_ids)}",
            _LE.DEBUG,
        )

        # Step 2: run pdb2gmx on the protein component only

        arguments_pdb2gmx = self._parse_arguments(
            flag_dict={
                "-f": os.path.join(tmp_dir, _SGE.PROTEIN_PDB),
                "-o": os.path.join(tmp_dir, _SGE.PROTEIN_PDB),
                "-p": _SGE.PROTEIN_TOP,
            }
        )
        self._backend_executor.execute(
            command=_GE.PDB2GMX, arguments=arguments_pdb2gmx, location=tmp_dir
        )
        # instantiate a separate topology object that will do a beter job of writing out than parmed can

        topol.forcefield = self.settings.arguments.parameters["-ff"]
        topol.water = self.settings.arguments.parameters["-water"]
        topol.parse(tmp_dir, _SGE.PROTEIN_TOP)
        topol.set_structure(tmp_dir, _SGE.PROTEIN_PDB, sanitize=True)
        posre_files = [
            f for f in os.listdir(tmp_dir) if f.endswith(".itp") and "posre" in f
        ]
        topol.add_posre(tmp_dir, posre_files)

        param_method = self._get_additional_setting(
            _SGE.PARAM_METHOD, default=_SGE.GAFF
        )
        for lig in lig_ids:
            input_file = lig + ".pdb"
            if param_method == _SGE.GAFF:
                # generate the itp files for each component, named by their PDB identifier
                self._generate_gaff_params(tmp_dir, input_file, topol)
            elif param_method == _SGE.OPENFF:
                self._generate_openff_params(tmp_dir, input_file, topol)
            else:
                raise NotImplementedError

        os.remove(os.path.join(tmp_dir, _SGE.PROTEIN_TOP))
        os.remove(os.path.join(tmp_dir, _SGE.PROTEIN_PDB))
        # final writeout of parametrized system
        topol.write_structure(tmp_dir, _SGE.COMPLEX_PDB)
        # convert pdb to gro
        editconf_arguments = ["-f", _SGE.COMPLEX_PDB, "-o", _SGE.STD_STRUCTURE]
        self._backend_executor.execute(
            command=_GE.EDITCONF,
            arguments=editconf_arguments,
            location=tmp_dir,
            check=True,
        )
        topol.set_structure(tmp_dir)
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



import numpy as np

from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR

from icolos.utils.enums.step_enums import (
    StepActiveLearningEnum,
)

_SALE = StepActiveLearningEnum()


class SurrogateModel:
    """class for the surrogate model used in prospective REINVENT"""

    # TODO: only random forests have been tested so far. Test SVR and maybe deep learning models
    def __init__(self, model_type: str):
        def initialize_RF():
            model = RandomForestRegressor(
                # enforcing max-depth may cause trees to be unable to predict extreme values
                # due to under-representation in the training data
                n_estimators=100, n_jobs=-1, criterion="mse"
            )

            return model

        def initialize_SVR():
            model = SVR(kernel="rbf", gamma="scale")
            return model

        self.type = model_type.lower()

        # random forest regressor
        if self.type == _SALE.RANDOM_FOREST_REGRESSOR:
            self.model = initialize_RF()
        # support vector regressor
        elif self.type == _SALE.SUPPORT_VECTOR_REGRESSOR:
            self.model = initialize_SVR()
        else:
            raise ValueError("Model not supported.")

    def fit(self, X_train, y_train):
        """train surrogate model from scratch. No online learning for random forest and support vectors"""
        if self.type == _SALE.RANDOM_FOREST_REGRESSOR:
            self.model.fit(X_train, y_train)
        if self.type == _SALE.SUPPORT_VECTOR_REGRESSOR:
            self.model.fit(np.array(X_train), y_train)

    def predict(self, data):
        """make predictions given query fingerprints"""
        if (
            self.type == _SALE.RANDOM_FOREST_REGRESSOR
            or self.type == _SALE.SUPPORT_VECTOR_REGRESSOR
        ):
            return self.model.predict(data)

    def get_std(self, X):
        """
        get standard deviation. This method is only implemented for random forests and is calculated
        using the prediction differences of each individual tree in the ensemble.
        """
        if self.type == _SALE.RANDOM_FOREST_REGRESSOR:
            individual_trees = self.model.estimators_
            subEstimates = np.array([tree.predict(X) for tree in individual_trees])
            return np.std(subEstimates, axis=0)

        elif self.type == _SALE.SUPPORT_VECTOR_REGRESSOR:
            raise NotImplementedError(
                "Standard deviation calculation not supported for SVR"
            )



from rdkit import Chem
from rdkit.Chem import rdmolops

from icolos.utils.enums.compound_enums import CompoundTagsEnum


def get_charge_for_molecule(molecule: Chem.Mol, add_as_tag=False) -> int:
    _MTE = CompoundTagsEnum()
    charge = rdmolops.GetFormalCharge(molecule)
    if add_as_tag:
        molecule.SetProp(_MTE.FORMAL_CHARGE_TAG, str(charge))
    return charge


def write_molecule_to_sdf(path: str, molecule: Chem.Mol):
    if molecule is None or not isinstance(molecule, Chem.Mol):
        raise ValueError("Function requires input attribute to be an RDkit molecule.")
    writer = Chem.SDWriter(path)
    writer.write(molecule)
    writer.close()



from icolos.config_containers.container import ConfContainer


class WorkflowContainer(ConfContainer):
    def __init__(self, conf, validation=True):
        super().__init__(conf=conf)

        # TODO: include validation with JSON Schema
        if validation:
            self.validate()

    def validate(self):
        pass



from typing import Dict, List, Optional
import uuid
from IPython.lib.display import IFrame
import pandas as pd
from icolos.core.containers.compound import Compound, Conformer
from pyvis.network import Network
from icolos.core.containers.generic import GenericData
from icolos.utils.enums.parallelization import ParallelizationEnum

from icolos.utils.enums.step_enums import StepFepPlusEnum
import os
from pydantic import BaseModel


_SFE = StepFepPlusEnum()
_PE = ParallelizationEnum


class Node(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    node_id: str = None
    node_hash: str = None
    conformer: Optional[Conformer] = None
    node_connectivity: List = []

    def __init__(self, **data) -> None:
        super().__init__(**data)

    def get_node_id(self) -> str:
        return self.node_id

    def get_node_color(self):
        # TODO: Expand this so we have different colours for each connectivity number [1,10]
        # this is just a placeholder for now
        thresholds = {i: "c0affe" for i in range(10)}

        num_connections = len(self.node_connectivity)
        return thresholds[num_connections]

    def set_node_id(self, node_id: str):
        self.node_id = node_id

    def get_conformer(self) -> Conformer:
        return self.conformer

    def set_conformer(self, conformer: Conformer) -> None:
        self.conformer = conformer

    def get_node_hash(self) -> str:
        return self.node_hash


class Edge(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    node_from: Node = Node()
    node_to: Node = Node()
    total: str = None
    mcs: str = None
    chg: str = None
    softbond: str = None
    min_no_atoms: str = None
    snapCoreRmsd: str = None
    bidirSnapCoreRmsd: str = None
    status: _PE = _PE.STATUS_SUCCESS
    ddG: float = 0.0
    ddG_err: float = 0.0

    def __init__(self, **data):
        super().__init__(**data)

    def _get_source_node_name(self):
        return self.node_from.get_node_hash()

    def _get_destination_node_name(self):
        return self.node_to.get_node_hash()

    def get_edge_id(self) -> str:
        # construct the edge ID from the node hashes, separated by '_'
        return f"{self.node_from.get_node_hash()}_{self.node_to.get_node_hash()}"

    def _set_status(self, status: str):
        assert status in [_PE.STATUS_SUCCESS, _PE.STATUS_FAILED]
        self.status = status


class PerturbationMap(BaseModel):
    """Hold a map construction parsed from a csv (probabably from a parsed schrodinger log
    file or something) and provide some utility methods for doing pmx calculations on the edges"""

    class Config:
        arbitrary_types_allowed = True

    nodes: List[Node] = []
    edges: List[Edge] = []
    hash_map: Dict = {}
    compounds: List[Compound] = []
    protein: GenericData = None
    vmap_output: IFrame = None
    replicas: int = 3
    node_df: pd.DataFrame = None
    # prune subsequent edge calculations on error
    strict_execution: str = False
    hub_conformer: Conformer = None

    def __init__(self, **data) -> None:
        super().__init__(**data)

    def _get_line_idx(self, data, id_str) -> int:
        line = [e for e in data if id_str in e]
        assert len(line) == 1
        line = line[0]
        return data.index(line)

    def _get_conformer_by_id(
        self, comp_name: str = None, comp_idx: int = None
    ) -> Conformer:
        """Parses the compound names from the edges specified in the mapper log file, performs a lookup to extract the corresponding compound from the step's compounds

        :param str comp_id: id of the compound parsed from the map generation
        :return Conformer: return the conformer object from self.data.compounds
        """
        for compound in self.compounds:
            # compound could have any arbitrary name - check whether this the node ID from schrodinger, or check the index string
            if (
                compound.get_name() == comp_name
                or compound.get_compound_number() == comp_idx
            ):
                # there will be one and only one conformer attached to one of the enumerations.  Return the first.
                for enum in compound.get_enumerations():
                    if enum.get_conformers():
                        return enum.get_conformers()[0]

    def generate_from_lomap_output(self, file: str):
        output_table = pd.read_csv(file, delimiter="   ,", index_col=False)
        output_table = output_table[output_table["Connect"].str.contains("Yes") == True]
        output_table.set_axis(
            [
                "Index_1",
                "Index_2",
                "Filename_1",
                "Filename_2",
                "Str_sim",
                "Eff_sim",
                "Loose_sim",
                "Connect",
            ],
            axis=1,
            inplace=True,
        )
        # generate a unique hash for each compound number, then we have mapping from compound number to node hash
        compound_df = pd.read_csv("out.txt", delimiter="\t", index_col="#ID")
        unique_hashes = [uuid.uuid4().hex for _ in compound_df.FileName]
        compound_df["NodeHash"] = unique_hashes

        # construct hash_ids for each edge
        hash_ids = []
        for node_from, node_to in zip(
            output_table["Filename_1"], output_table["Filename_2"]
        ):
            node_from = node_from.strip()
            node_to = node_to.strip()
            hash_from = compound_df[
                compound_df["FileName"] == node_from
            ].NodeHash.values[0]
            hash_to = compound_df[compound_df["FileName"] == node_to].NodeHash.values[0]
            hash_ids.append(f"{hash_from}_{hash_to}")
        output_table["EdgeHash"] = hash_ids

        for _, row in compound_df.iterrows():
            compound_id = row.FileName.split(".")[0].split(":")[0]
            self.nodes.append(
                Node(
                    node_id=compound_id,
                    node_hash=row.NodeHash,
                    conformer=self._get_conformer_by_id(comp_idx=int(compound_id)),
                )
            )

        for _, edge in output_table.iterrows():
            edge = Edge(
                node_from=self._get_node_by_hash_id(edge.EdgeHash.split("_")[0]),
                node_to=self._get_node_by_hash_id(edge.EdgeHash.split("_")[1]),
                mcs=float(edge.Str_sim),
            )
            self.edges.append(edge)

        for node in self.nodes:
            self._attach_node_connectivity(node)

    def generate_star_map(self) -> None:
        """Generates a star topology using a single hub compound"""
        hub_node = Node(
            node_id=self.hub_compound.get_index_string(),
            node_hash=uuid.uuid4().hex,
            conformer=self.hub_compound.get_molecule(),
        )
        for compound in self.compounds:
            end_node = Node(
                node_id=compound.get_index_string(),
                node_hash=uuid.uuid4().hex,
                conformer=compound.get_enumerations()[0]
                .get_conformers()[0]
                .get_molecule(),
            )
            edge = Edge(
                node_from=hub_node,
                node_to=end_node,
            )
            self.edges.append(edge)

        for node in self.nodes:
            self._attach_node_connectivity(node)

    def parse_map_file(self, file_path: str) -> None:
        """Parse map from Schrodinger's fep_mapper log file, build internal graph representation + attach properties from fmp_stats, if provided

        :param str file_path: path to the fep_mapper.log file to extract the perturbation map from
        """
        # we need to do some format enforcement here (schrodinger or otherwise)

        with open(file_path, "r") as f:
            data = f.readlines()

        start_node = self._get_line_idx(data, _SFE.NODE_HEADER_LINE)
        stop_node = self._get_line_idx(data, _SFE.SIMULATION_PROTOCOL)
        edge_info_start = self._get_line_idx(data, _SFE.SIMILARITY)

        # TODO: refactor that
        # clean up the data from schrodinger
        split_data = []
        for line in data:
            split_line = line.split("  ")
            stripped_line = []
            for element in split_line:
                if not element.isspace() and element:
                    stripped_line.append(element.strip())
            split_data.append(stripped_line)

        data = split_data

        self.node_df = pd.DataFrame(
            data[start_node + 3 : stop_node - 1],
            index=None,
            columns=[
                "hash_id",
                "node_id",
                "Predicted dG",
                "Experimental dG",
                "Predicted Solvation dG",
                "Experimental Solvation dG",
            ],
        )
        edge_info = pd.DataFrame(
            data[edge_info_start + 3 : -1],
            columns=[
                "Short ID",
                "Total",
                "Mcs",
                "Charge",
                "SoftBond",
                "MinimumNumberOfAtom",
                "SnapCoreRmsd",
                "BidirectionSnapCore",
            ],
        ).dropna()
        for hash_id, node_id in zip(self.node_df["hash_id"], self.node_df["node_id"]):
            # map the hashes to the compound IDs
            self.hash_map[hash_id] = node_id
            node = Node(
                node_id=node_id,
                node_hash=hash_id,
                conformer=self._get_conformer_by_id(node_id),
            )
            # generate the Node object to wrap the compound
            self.nodes.append(node)

        for _, edge in edge_info.iterrows():
            edge = Edge(
                node_from=self._get_node_by_hash_id(edge[0].split("_")[0]),
                node_to=self._get_node_by_hash_id(edge[0].split("_")[1]),
                total=edge[1],
                mcs=edge[2],
                chg=edge[3],
                softbond=edge[4],
                min_no_atoms=edge[5],
                snapCoreRmsd=edge[6],
                bidirSnapCoreRmsd=edge[7],
            )
            self.edges.append(edge)
        # process the node info, generate the hash map
        for node in self.nodes:
            self._attach_node_connectivity(node)

    def _attach_node_connectivity(self, node: Node):
        # looks through the constructed edges, returns ids of any edges that have the specified node as one component
        connected_edges = []
        for edge in self.edges:
            if (
                edge.node_from.get_node_hash() == node.node_hash
                or edge.node_to.get_node_hash() == node.node_hash
            ):
                connected_edges.append(edge.get_edge_id())
        node.node_connectivity = connected_edges

    def _get_node_by_node_id(self, node_id: str) -> Node:
        for node in self.nodes:
            if node.node_id == node_id:
                return node

    def _get_node_by_hash_id(self, hash_id: str) -> Node:
        for node in self.nodes:
            if node.node_hash == hash_id:
                return node

    def get_edges(self, alive_only=True) -> List[Edge]:
        if alive_only:
            return [e for e in self.edges if e.status == _PE.STATUS_SUCCESS]
        else:
            return self.edges

    def get_nodes(self) -> List[Node]:
        return self.nodes

    def add_edge(self, edge: Edge) -> None:
        self.edges.append(edge)

    def add_node(self, node: Node) -> None:
        self.nodes.append(node)

    def visualise_perturbation_map(self, write_out_path: str) -> None:
        """Generate NetworkX graph for the parsed perturbation map

        :param str write_out_path: directory to write output file
        """
        vmap = Network(directed=True)
        vmap.barnes_hut()

        # this is not an iterable
        for edge in self.edges:
            vmap.add_node(
                edge._get_source_node_name(), color=edge.node_from.get_node_color()
            )
            vmap.add_node(
                edge._get_destination_node_name(), color=edge.node_to.get_node_color()
            )
            vmap.add_edge(
                source=edge._get_source_node_name(),
                to=edge._get_destination_node_name(),
                length=edge.total,
                label="total: " + str(edge.total),
                title="Mcs: " + str(edge.mcs) + ", SnapCoreRMSD: ",
            )
        self.vmap_output = vmap.show(os.path.join(write_out_path, "vmap.html"))
        # return self.vmap_output

    def get_protein(self) -> GenericData:
        return self.protein

    def get_edge_by_id(self, id: str) -> Optional[Edge]:
        """Lookup edge by identifier

        :param str id: edge hash to retrieve
        :return Optional[Edge]: Return the edge if found, else None
        """
        # handle case where the task is actually a path to a batch script
        if not isinstance(id, Edge):
            parts = id.split("/")

            for part in parts:
                for e in self.edges:
                    if part == e:
                        id = e

        match = [e for e in self.edges if e.get_edge_id() == id]
        if not match:
            return
        else:
            return match[0]

    def __repr__(self) -> str:
        return f"Icolos Perturbation Map object containing {len(self.edges)} edges and {len(self.nodes)} nodes"

    def __str__(self) -> str:
        return self.__repr__()



from icolos.utils.enums.program_parameters import AutoDockVinaEnum
from icolos.utils.execute_external.execute import ExecutorBase

_EE = AutoDockVinaEnum()


class AutoDockVinaExecutor(ExecutorBase):
    """For the execution of AutoDock Vina 1.2.0."""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided
        if command not in [_EE.VINA_CALL]:
            raise ValueError(
                "Parameter command must be in dictionary of the internal AutoDock Vina executable list."
            )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=None,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(
                command=_EE.VINA_CALL, arguments=[_EE.VINA_HELP], check=True
            )
            if result.returncode == 0:
                return True
            return False
        except Exception as e:
            return False



from typing import Dict, List
from icolos.core.containers.perturbation_map import Edge
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
import os
from icolos.utils.enums.program_parameters import (
    GromacsEnum,
    StepPMXEnum,
)
from icolos.utils.execute_external.pmx import PMXExecutor
from icolos.utils.general.parallelization import SubtaskContainer

_PSE = StepPMXEnum()
_GE = GromacsEnum()


class StepPMXBoxWaterIons(StepPMXBase, BaseModel):
    """
    Take the prepared structure files and prepare the system,
    runs editconf, solvate, genion and grompp for each system
    to be simulated
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=PMXExecutor)

    def execute(self):

        edges = [e.get_edge_id() for e in self.get_edges()]

        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self.boxWaterIons,
            step_id="pmx boxWaterIons",
            result_checker=self._check_result,
        )

    def boxWaterIons(self, jobs: List[str]):
        mdp_path = os.path.join(self.work_dir, "input/mdp")

        for edge in jobs:
            outLigPath = self._get_specific_path(
                workPath=self.work_dir, edge=edge, wp="unbound"
            )
            outProtPath = self._get_specific_path(
                workPath=self.work_dir, edge=edge, wp="bound"
            )

            # box ligand
            inStr = "{0}/init.pdb".format(outLigPath)
            outStr = "{0}/box.pdb".format(outLigPath)
            editconf_args = [
                "-f",
                inStr,
                "-o",
                outStr,
                "-bt",
                self.boxshape,
                "-d",
                self.boxd,
            ]
            self._gromacs_executor.execute(
                command=_GE.EDITCONF, arguments=editconf_args, check=True
            )
            # box protein
            inStr = "{0}/init.pdb".format(outProtPath)
            outStr = "{0}/box.pdb".format(outProtPath)
            editconf_args = [
                "-f",
                inStr,
                "-o",
                outStr,
                "-bt",
                self.boxshape,
                "-d",
                self.boxd,
            ]
            self._gromacs_executor.execute(
                command=_GE.EDITCONF, arguments=editconf_args, check=True
            )
            # water ligand
            inStr = "{0}/box.pdb".format(outLigPath)
            outStr = "{0}/water.pdb".format(outLigPath)
            top = "{0}/topol.top".format(outLigPath)
            solvate_args = ["-cp", inStr, "-cs", "spc216.gro", "-p", top, "-o", outStr]
            self._gromacs_executor.execute(
                command=_GE.SOLVATE,
                arguments=solvate_args,
                check=True,
                location=self.work_dir,
            )
            # water protein
            inStr = "{0}/box.pdb".format(outProtPath)
            outStr = "{0}/water.pdb".format(outProtPath)
            top = "{0}/topol.top".format(outProtPath)
            solvate_args = ["-cp", inStr, "-cs", "spc216.gro", "-p", top, "-o", outStr]
            self._gromacs_executor.execute(
                command=_GE.SOLVATE, arguments=solvate_args, location=self.work_dir
            )

            # ions ligand
            inStr = "{0}/water.pdb".format(outLigPath)
            outStr = "{0}/ions.pdb".format(outLigPath)
            mdp = "{0}/em_l0.mdp".format(mdp_path)
            tpr = "{0}/tpr.tpr".format(outLigPath)
            top = "{0}/topol.top".format(outLigPath)
            mdout = "{0}/mdout.mdp".format(outLigPath)
            grompp_args = [
                "-f",
                mdp,
                "-c",
                inStr,
                "-r",
                inStr,
                "-p",
                top,
                "-o",
                tpr,
                "-maxwarn",
                4,
                "-po",
                mdout,
            ]

            self._gromacs_executor.execute(
                command=_GE.GROMPP,
                arguments=grompp_args,
                check=True,
                location=self.work_dir,
            )
            genion_args = [
                "-s",
                tpr,
                "-p",
                top,
                "-o",
                outStr,
                "-conc",
                self.conc,
                "-pname",
                self.pname,
                "-nname",
                self.nname,
                "-neutral",
            ]
            self._gromacs_executor.execute(
                command=_GE.GENION,
                arguments=genion_args,
                check=True,
                pipe_input="echo SOL",
                location=self.work_dir,
            )
            # ions protein
            inStr = "{0}/water.pdb".format(outProtPath)
            outStr = "{0}/ions.pdb".format(outProtPath)
            mdp = "{0}/em_l0.mdp".format(mdp_path)
            tpr = "{0}/tpr.tpr".format(outProtPath)
            top = "{0}/topol.top".format(outProtPath)
            mdout = "{0}/mdout.mdp".format(outProtPath)
            grompp_args = [
                "-f",
                mdp,
                "-c",
                inStr,
                "-r",
                inStr,
                "-p",
                top,
                "-o",
                tpr,
                "-maxwarn",
                4,
                "-po",
                mdout,
            ]

            self._gromacs_executor.execute(
                command=_GE.GROMPP,
                arguments=grompp_args,
                check=True,
                location=self.work_dir,
            )
            genion_args = [
                "-s",
                tpr,
                "-p",
                top,
                "-o",
                outStr,
                "-conc",
                self.conc,
                "-pname",
                self.pname,
                "-nname",
                self.nname,
                "-neutral",
            ]
            self._gromacs_executor.execute(
                command=_GE.GENION,
                arguments=genion_args,
                check=True,
                pipe_input="echo SOL",
                location=self.work_dir,
            )

            # clean backed files
            self._clean_backup_files(outLigPath)

            self._clean_backup_files(outProtPath)

    def _check_result(self, batch: List[List[str]]) -> List[List[bool]]:
        """
        Look in each hybridStrTop dir and check the output pdb files exist for the edges
        """
        output_files = [f"{f}/tpr.tpr" for f in self.therm_cycle_branches]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(os.path.join(self.work_dir, job, f))
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results



import argparse


def str2bool(inp):
    if isinstance(inp, bool):
        return inp
    if inp.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif inp.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError(
            "Expected castable string or boolean value as input."
        )



from icolos.core.workflow_steps.schrodinger.fep_base import StepFEPBase
from icolos.utils.enums.step_enums import StepBaseEnum, StepFepPlusEnum
from icolos.utils.enums.program_parameters import FepPlusEnum

from pydantic import BaseModel

_FE = FepPlusEnum()
_SFE = StepFepPlusEnum()
_SBE = StepBaseEnum


class StepFepPlusAnalysis(StepFEPBase, BaseModel):
    """
    Standalone class to analyse data from a previous fep job
    """

    def __init__(self, **data):
        super().__init__(**data)

    def execute(self):
        """
        Analyses the map produced from an FEP run
        """
        tmp_dir = self._prepare_tmpdir()
        self.data.generic.write_out_all_files(tmp_dir)
        self._extract_log_file_data(tmp_dir)
        self._remove_temporary(tmp_dir)



from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.enums.step_enums import StepGromacsEnum
from typing import List
from pydantic import BaseModel
from icolos.utils.execute_external.execute import Executor
from icolos.core.workflow_steps.step import _LE
from icolos.utils.enums.program_parameters import GromacsEnum
import os
import sys

_SGE = StepGromacsEnum()
_GE = GromacsEnum()


class StepClusterTS(StepGromacsBase, BaseModel):
    """
    Generate time-resolved cluster plots from the output of gmx cluster, relies on MDplot R package
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=Executor)

    def _construct_args(self, defaults: dict) -> List:
        """
        Custom method for argument construction, includes checks for required args in the config.
        """
        args = []
        for key, value in self.settings.arguments.parameters.items():
            args.append("".join([key, "=", value]))

        for value in self.settings.arguments.flags:
            args.append(value)
        for key, value in defaults.items():
            if key not in self.settings.arguments.parameters.keys():
                args.append("".join([key, "=", value]))

        # do some checks to make sure the required params have been passed
        for arg in [_SGE.CLUSTERS_NUMBER, _SGE.LENGTHS]:
            if arg not in self.settings.arguments.parameters.keys():
                self._logger.log(
                    f"Argument for parameter {arg} not found in provided argument. \
                        This must be specified!.  If this workflow has attached stdin, \
                            you can enter the value now...",
                    _LE.WARNING,
                )
                # instead of bailing out, take input from user if process has stdin connected
                if sys.stdin and sys.stdin.isatty():
                    value = input(f"Provide the parameter for option {arg}>>>")
                    args.append("".join([key, "=", value]))
                else:
                    self._logger.log(
                        f"No stdin stream detected, and cannot infer argument, step {self.step_id} may fail",
                        _LE.WARNING,
                    )
        return args

    def execute(self):
        """
        Visualise time-resolved gmx cluster results.
        Requires predceeding gmx_cluster step with clust-id.xvg file
        (ensure -clid flag is set, and xvg file is passed to this step)
        """

        tmp_dir = self._make_tmpdir()
        self.data.generic.write_out_all_files(tmp_dir)
        xvg_file = self.data.generic.get_argument_by_extension(ext="xvg")

        arguments = self._construct_args(
            defaults={
                "files": os.path.join(tmp_dir, xvg_file),
                "size": "1500,1500",
                "outformat": "png",
                "outfile": "clusters_ts.png",
                "timeUnit": "ns",
                "title": "CLUSTERS_timeseries",
            },
        )

        self._backend_executor.execute(
            command=_GE.CLUSTER_TS, arguments=arguments, location=tmp_dir, check=True
        )

        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



from pydantic import BaseModel
from icolos.core.containers.compound import Compound
from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path
from icolos.utils.enums.step_enums import StepFepPlusEnum
from typing import List
import time
import os
from icolos.core.workflow_steps.step import _LE

_SFE = StepFepPlusEnum()


class StepFEPBase(StepSchrodingerBase, BaseModel):
    """
    Base class containing common functionality for Schrodinger FEP+ workflows
    """

    def __init__(self, **data):
        super().__init__(**data)

    def _parse_output(self, tmp_dir):
        # pick up the final annotated map construction
        self.data.generic.clear_file_dict()
        self._logger.log(f"Reading output map.", _LE.INFO)
        data = None
        counts = 0
        # hold whilst the job data gets written to local fs
        while data is None and counts < 50000:
            try:
                path = [
                    file
                    for file in os.listdir(tmp_dir)
                    if file.endswith(_SFE.FMP_OUTPUT_FILE)
                ]
                assert len(path) == 1
                path = path[0]
                with open(os.path.join(tmp_dir, path), "rb") as f:
                    data = f.read()
            except AssertionError:
                self._logger.log(
                    "Output file has not yet appeared in the file system, sleeping and retrying...",
                    _LE.INFO,
                )
                time.sleep(15)
                counts += 1

        self._add_data_to_generic(path, data)

    def _extract_log_file_data(self, tmp_dir):
        """
        Parses FEP log file to extract edge and node properties
        """
        lines = None
        counts = 0
        # wait whilst job sits in the queue
        while lines is None and counts < 50000:
            try:
                log_file = [
                    file for file in os.listdir(tmp_dir) if file.endswith(_SFE.LOGFILE)
                ]
                assert len(log_file) == 1
                log_file = log_file[0]

                with open(os.path.join(tmp_dir, log_file), "r") as f:
                    lines = f.readlines()

                edge_header_index = [
                    idx for idx, s in enumerate(lines) if _SFE.EDGE_HEADER_LINE in s
                ][-1]
                node_header_index = [
                    idx for idx, s in enumerate(lines) if _SFE.NODE_HEADER_LINE in s
                ][-1]
                end_of_data_index = [
                    idx for idx, s in enumerate(lines) if _SFE.DATA_TERMINUS in s
                ][0]

                edge_data_lines = [
                    line
                    for line in lines[edge_header_index + 3 : node_header_index - 1]
                ]
                node_data_lines = [
                    line
                    for line in lines[node_header_index + 3 : end_of_data_index - 1]
                ]

                self._process_edge_lines(edge_data_lines)
                self._process_node_lines(node_data_lines)

            except AssertionError:
                self._logger.log(
                    "Log file has not yet appeared in the file system, sleeping and retrying...",
                    _LE.INFO,
                )
                time.sleep(15)
                counts += 1

    def _process_node_lines(self, data: List[str]) -> None:
        for entry in data:
            fields = entry.split()
            idx = fields[1]
            dG = fields[2]
            # attach dG tags to compound objects if present
            if self.data.compounds:
                # account for running this step compoundless
                self.data.compounds[int(idx[0])].get_enumerations()[0].get_conformers()[
                    0
                ].get_molecule().SetProp("dG", str(dG))
            self._logger.log(
                f"dG directly from the output file for compound {idx} is {dG} ",
                _LE.INFO,
            )

    def _process_edge_lines(self, edge_data: List[str]) -> None:
        """
        Calibrate dG values using a reference compound and edge ddG from log file output, return dG for each compound
        """

        # caluclate the max ligand index, accounting for ligands that may have been skipped in previous steps, so can't rely on self.get_compounds()
        len_nodes = 0
        for line in edge_data:
            parts = line.split()

            lig_from = int(parts[1].split(":")[0])
            lig_to = int(parts[3].split(":")[0])
            for idx in [lig_from, lig_to]:
                if idx > len_nodes:
                    len_nodes = idx
        len_nodes += 1  # account for zero indexed ligands

        error_matrix = np.zeros((len_nodes, len_nodes))
        ddG_matrix = np.zeros((len_nodes, len_nodes))
        for line in edge_data:
            parts = line.split()
            try:
                # parse the compound info from the log file
                lig_from = int(parts[1].split(":")[0])
                lig_to = int(parts[3].split(":")[0])
                ddG = float(parts[4].split("+-")[0])
                err = float(parts[4].split("+-")[1])
            except ValueError:
                self._logger.log(
                    f"Line: {line} from the logfile contained an unexpected datatype - cannot process this edge - skipping",
                    _LE.WARNING,
                )
                continue

            error_matrix[lig_from, lig_to] = err
            error_matrix[lig_to, lig_from] = err
            ddG_matrix[lig_from, lig_to] = ddG
            ddG_matrix[lig_to, lig_from] = -ddG
        error_matrix = csr_matrix(error_matrix)
        # compute shortest path from one ligand to the anchor
        _, predecessors = shortest_path(
            error_matrix, directed=False, return_predecessors=True, indices=0
        )
        self._construct_dg_per_compound(
            ddG_matrix, predecessors, error_matrix, self.get_compounds()
        )

    def _construct_dg_per_compound(
        self,
        ddG: np.ndarray,
        predecessors: List,
        error_matrix: np.ndarray,
        compounds: List[Compound],
    ) -> None:
        """
        Calculate the calibrated binding free energy per compound using a reference value
        Attach calcualted dG to compounds
        """
        try:
            ref_dG = self.settings.additional[_SFE.REFERENCE_DG]
        except KeyError:
            self._logger.log(
                "Expected to find a reference dG value for the lead compound, but none was found."
                "Defaulting to 0.00, you will need to apply a manual correction afterwards",
                _LE.WARNING,
            )
            ref_dG = 0.00

        def _calculate_dg(comp_num: int, dG=ref_dG, err=0):
            prev_index = predecessors[comp_num]
            dG += ddG[prev_index, comp_num]
            err += error_matrix[prev_index, comp_num]
            if prev_index != 0:
                _calculate_dg(prev_index, dG=dG, err=err)
            else:
                data = str(round(dG, 2)) + "+-" + str(round(err, 2))
                compounds[idx].get_enumerations()[0].get_conformers()[
                    0
                ].get_molecule().SetProp("map_dG", data)
                self._logger.log(
                    f"Calculated dG from spanning tree for compound {idx} is {data}",
                    _LE.INFO,
                )

        for comp in compounds:
            idx = comp.get_compound_number()
            # check whether the compound appeared in the final map
            try:

                if idx == 0:
                    comp.get_enumerations()[0].get_conformers()[
                        0
                    ].get_molecule().SetProp(
                        "map_dG", str(self.settings.additional[_SFE.REFERENCE_DG])
                    )
                if idx != 0:  # skip the reference compound
                    _calculate_dg(idx)
            except IndexError:
                self._logger.log(
                    f"Compound {idx} was not found in the output map, it was likely dropped during the workflow",
                    _LE.WARNING,
                )
                continue



from pydantic import BaseModel, PrivateAttr

from icolos.core.composite_agents.base_agent import BaseAgent, AgentHeaderParameters

from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.composite_agents_enums import SchedulerEnum

_SE = SchedulerEnum()
_LE = LoggingConfigEnum()


class SchedulerHeaderParameters(AgentHeaderParameters, BaseModel):
    pass


class Scheduler(BaseAgent, BaseModel):
    """Class to hold the whole logic for scheduling sub-jobs."""

    header: SchedulerHeaderParameters = SchedulerHeaderParameters()

    class Config:
        underscore_attrs_are_private = True

    _logger = PrivateAttr()
    _initialized_steps = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

    def initialize(self):
        super().initialize()

    def execute(self):
        # TODO: implement
        pass

    def _action_prepare(self):
        pass

    def _action_run(self):
        pass

    def is_valid(self) -> bool:
        # TODO: implement
        pass

    def __repr__(self):
        return "<Icolos scheduler: id=%s, description=%s>" % (
            self.get_id(),
            self.get_description(),
        )

    def __str__(self):
        return self.__repr__()



from typing import Dict, List
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
import os
from icolos.utils.enums.program_parameters import PMXAtomMappingEnum, PMXEnum
from icolos.utils.execute_external.pmx import PMXExecutor
from icolos.utils.general.parallelization import SubtaskContainer

_PE = PMXEnum()
_PAE = PMXAtomMappingEnum()


class StepPMXAssembleSystems(StepPMXBase, BaseModel):
    """
    Executes the assemble_systems.py script, edges are parallelized over available cores

    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=PMXExecutor)

    def execute(self):
        assert self.work_dir is not None and os.path.isdir(self.work_dir)

        # get edges from the perturbation map attached to the step
        edges = [e.get_edge_id() for e in self.get_edges()]

        # enforce one edge per task list (results in multiple batches for large maps)
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self._execute_command,
            step_id="pmx assemble_systems",
            result_checker=self._check_results,
        )

    def _execute_command(self, jobs: List):

        args = {
            "-edges": '"' + " ".join([e for e in jobs]) + '"',
            "-ligand_path": os.path.join(self.work_dir, _PAE.LIGAND_DIR),
            "-workPath": self.work_dir,
        }
        self._backend_executor.execute(
            command=_PE.ASSEMBLE_SYSTEMS,
            arguments=self.get_arguments(defaults=args),
            check=True,
            location=self.work_dir,
        )

    def _check_results(self, batch: List[List[str]]) -> List[List[bool]]:
        output_files = ["bound/init.pdb", "unbound/init.pdb"]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(os.path.join(self.work_dir, job, f))
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results



from typing import List
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import numpy as np
from icolos.utils.enums.execution_enums import ExecutionPlatformEnum
from icolos.utils.enums.program_parameters import (
    StepPMXEnum,
)
from icolos.utils.execute_external.slurm_executor import SlurmExecutor
from icolos.utils.general.parallelization import Subtask, SubtaskContainer
import os

_PSE = StepPMXEnum()
_EPE = ExecutionPlatformEnum


class StepPMXRunSimulations(StepPMXBase, BaseModel):
    """
    Runs md simulations, unwraps into pool of individual jobs, parallelized over available GPUs
    """

    sim_type: str = ""

    def __init__(self, **data):
        super().__init__(**data)

        # Note: if youre running the job on, for example, a workstation, without slurm, this will simply execute the scripts directly (the slurm header is simply ignored in this case)
        self._initialize_backend(executor=SlurmExecutor)

    def execute(self):

        if self.run_type == "rbfe":
            edges = [e.get_edge_id() for e in self.get_edges()]
        elif self.run_type == "abfe":
            edges = [c.get_index_string() for c in self.get_compounds()]
        self.sim_type = self._get_additional_setting(_PSE.SIM_TYPE)
        assert (
            self.sim_type in self.mdp_prefixes.keys()
        ), f"sim type {self.sim_type} not recognised!"
        # prepare and pool jobscripts, unroll replicas,  etc
        job_pool = self._prepare_job_pool(edges)
        self._logger.log(
            f"Prepared {len(job_pool)} jobs for {self.sim_type} simulations",
            _LE.DEBUG,
        )

        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(job_pool)
        result_checker = (
            self._inspect_dhdl_files
            if self.sim_type == "transitions"
            else self._inspect_log_files
        )
        # simulations are run without the normal parallelizer
        # this relies upon job IDs from Slurm
        if (
            self.execution.platform == _EPE.SLURM
            and self._backend_executor.is_available()
        ):
            self._run_job_pool(run_func=self._run_single_job)
        else:
            # simulations are run locally
            self._execute_pmx_step_parallel(
                run_func=self._execute_command,
                step_id="pmx_run_simulations",
                # for run_simulations, because batch efficiency is crucial, we do this prior to batching
                prune_completed=False,
                result_checker=result_checker,
            )
        # clean up large files from completed transition jobs only
        # if self.sim_type == "transitions":
        #     self._logger.log("Cleaning up transition output files", _LE.DEBUG)
        #     to_clean_list = []
        #     for job in job_pool:
        #         # search branch, state, run
        #         to_clean_list.append(
        #             glob(f"{self.work_dir}/{job}/*/*/*/transitions/frame*.gro")
        #         )
        #     for file in to_clean_list:
        #         os.remove(file)

    def get_mdrun_command(
        self,
        tpr=None,
        ener=None,
        confout=None,
        mdlog=None,
        trr=None,
    ):

        mdrun_binary = self._get_additional_setting(
            _PSE.MDRUN_EXECUTABLE, default="gmx mdrun"
        )
        if self.sim_type in ("em", "eq", "npt", "nvt"):
            # add some logic to each commands to handle restarts
            mdrun_call = [
                mdrun_binary,
                "-s",
                tpr,
                "-e",
                ener,
                "-c",
                confout,
                "-o",
                trr,
                "-g",
                mdlog,
            ]
            for flag in self.settings.arguments.flags:
                mdrun_call.append(str(flag))
            for key, value in self.settings.arguments.parameters.items():
                mdrun_call.append(str(key))
                mdrun_call.append(str(value))

            job_command = (
                [f"FILE={os.path.dirname(tpr)}/state.cpt\nif [ -f $FILE ]; then\n"]
                + mdrun_call
                + ["-cpi state.cpt\nelse\n"]
                + mdrun_call
                + ["\nfi\n"]
            )

        elif self.sim_type == "transitions":
            # need to add many job commands to the slurm file, one for each transition
            sim_path = os.path.dirname(tpr)
            tpr_files = [f for f in os.listdir(sim_path) if f.endswith("tpr")]
            job_command = []
            # note, these will not be returned in numerical order!
            for file in tpr_files:
                # grab the index in the tpr file
                tpr_idx = file.split(".tpr")[0][2:]
                dhdl_file = os.path.join(os.path.dirname(mdlog), f"dhdl{tpr_idx}.xvg")
                if not os.path.isfile(dhdl_file):
                    # handles aws spot restarts
                    single_command = [
                        f"FILE={sim_path}/dhdl{tpr_idx}.xvg\nif [ ! -f $FILE ]; then\n"
                    ]
                    single_command += [
                        mdrun_binary,
                        "-s",
                        file,
                        "-e",
                        ener,
                        "-c",
                        confout,
                        "-dhdl",
                        f"dhdl{tpr_idx}.xvg",
                        "-o",
                        trr,
                        "-g",
                        mdlog,
                    ]
                    for flag in self.settings.arguments.flags:
                        single_command.append(str(flag))
                    for key, value in self.settings.arguments.parameters.items():
                        single_command.append(str(key))
                        single_command.append(str(value))

                    single_command.append("\nfi\n")
                    single_command.append(f"\nrm {os.path.dirname(ener)}/*#\n")
                    job_command += single_command
                else:
                    self._logger.log(
                        f"dhdl file for transition {tpr_idx} in {os.path.dirname(mdlog)} already exists, skipping",
                        _LE.DEBUG,
                    )
        return job_command

    def _prepare_single_job(self, edge: str, wp: str, state: str, r: int):
        """
        Construct a slurm job file in that job's directory, return the path to the batch script
        """
        simpath = self._get_specific_path(
            workPath=self.work_dir,
            edge=edge,
            wp=wp,
            state=state,
            r=r,
            sim=self.sim_type,
        )
        tpr = "{0}/tpr.tpr".format(simpath)
        ener = "{0}/ener.edr".format(simpath)
        confout = "{0}/confout.gro".format(simpath)
        mdlog = "{0}/md.log".format(simpath)
        trr = "{0}/traj.trr".format(simpath)
        if self.sim_type != "transitions":
            try:
                with open(os.path.join(simpath, "md.log"), "r") as f:
                    lines = f.readlines()
                sim_complete = any(["Finished mdrun" in l for l in lines])
            except FileNotFoundError:
                sim_complete = False

        # handle transitions
        else:
            # cannot reliably check that all sims for all edges have completed here, this will be checked in get_mdrun_command which will skip completed perturbations if dhdl exists
            sim_complete = False

        if not sim_complete:
            self._logger.log(
                f"Preparing: {wp} {edge} {state} run{r}, simType {self.sim_type}",
                _LE.DEBUG,
            )
            job_command = self.get_mdrun_command(
                tpr=tpr,
                trr=trr,
                ener=ener,
                confout=confout,
                mdlog=mdlog,
            )
            # empty list indicates all transitions completed
            if job_command:
                job_command = " ".join(job_command)
                batch_file = self._backend_executor.prepare_batch_script(
                    job_command, arguments=[], location=simpath
                )
                return os.path.join(simpath, batch_file)

        return

    def _prepare_job_pool(self, edges: List[str]):
        replicas = (
            self.get_perturbation_map().replicas
            if self.get_perturbation_map() is not None
            else 1
        )
        batch_script_paths = []
        # load in the protein jobs first, queue is FIFO
        for branch in self.therm_cycle_branches:
            for edge in edges:
                for r in range(1, replicas + 1):
                    for state in self.states:
                        path = self._prepare_single_job(
                            edge=edge, wp=branch, state=state, r=r
                        )
                        if path is not None:
                            batch_script_paths.append(path)
        return batch_script_paths

    def _run_single_job(self, job: str):
        """
        Execute the simulation for a single batch script
        """
        self._logger.log(
            f"Starting execution of job {job.split('/')[-1]}.",
            _LE.DEBUG,
        )
        location = os.path.dirname(job)
        job_id = self._backend_executor.execute(
            tmpfile=job, location=location, check=False, wait=False
        )
        return job_id

    def _execute_command(self, jobs: List[Subtask]):
        for idx, job in enumerate(jobs):
            self._logger.log(
                f"Starting execution of job {job.split('/')[-1]}. ",
                _LE.DEBUG,
            )
            self._logger.log(f"Batch progress: {idx+1}/{len(jobs)}", _LE.DEBUG)
            location = os.path.dirname(job)
            result = self._backend_executor.execute(
                tmpfile=job, location=location, check=False
            )
            self._logger.log(
                f"Execution for job {job} completed with status: {result}", _LE.DEBUG
            )

    def _inspect_log_files(self, jobs: List[str]) -> List[List[bool]]:
        """
        Check the md.log files in the edge's job dir, return
        :param jobs: list of paths to the batch scripts
        """
        results = []
        for subtask in jobs:
            subtask_results = []
            for sim in subtask:
                location = os.path.join(os.path.dirname(sim), "md.log")
                if os.path.isfile(location):
                    with open(location, "r") as f:
                        lines = f.readlines()
                    subtask_results.append(
                        any(["Finished mdrun" in l for l in lines[-20:]])
                    )
                else:
                    subtask_results.append(False)
            results.append(subtask_results)
        return results

    def _inspect_dhdl_files(self, jobs: List[str]) -> List[List[bool]]:
        # check there is a dhdl file for each tpr file
        results = []
        for subtask in jobs:
            subtask_results = []
            for sim in subtask:

                location = os.path.join(os.path.dirname(sim))
                n_snapshots = len(
                    [f for f in os.listdir(location) if f.endswith("tpr")]
                )
                dhdl_files = [f"dhdl{i}.xvg" for i in range(1, n_snapshots + 1)]

                if not all(
                    [os.path.isfile(os.path.join(location, f)) for f in dhdl_files]
                ):
                    subtask_results.append(False)
                else:
                    subtask_results.append(True)

            results.append(subtask_results)
        return results



from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

_GE = GromacsEnum()


class StepGMXTrjcat(StepGromacsBase, BaseModel):
    """
    Concatenates multiple trajectories, useful for subsequent rmsd/cluster calculations
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):

        tmp_dir = self._make_tmpdir()

        # write the trajectories to the tmpdir, writing to separate file names, then glob the xtc files

        for idx, file in enumerate(self.data.generic.get_files_by_extension(ext="xtc")):
            file.write(path=os.path.join(tmp_dir, f"traj_{idx}.xtc"), join=False)

        flag_dict = {
            "-f": "*.xtc",
            "-o": "trjcat_out.xtc",
            "-cat": "",  # need this to paste the trajectories back to back
        }

        arguments = self._parse_arguments(flag_dict=flag_dict)
        result = self._backend_executor.execute(
            command=_GE.TRJCAT, arguments=arguments, location=tmp_dir, check=True
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)

        rm_files = [
            f for f in os.listdir(tmp_dir) if f.endswith("xtc") and "trjcat" not in f
        ]
        for f in rm_files:
            os.remove(os.path.join(tmp_dir, f))
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



import lomap
from icolos.core.containers.perturbation_map import PerturbationMap
from icolos.core.workflow_steps.step import StepBase
from pydantic import BaseModel
import networkx as nx
import os
from icolos.utils.enums.program_parameters import OpenBabelEnum

from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from rdkit import Chem

_OBE = OpenBabelEnum()


class StepLomap(StepBase, BaseModel):
    _openbabel_executor: OpenBabelExecutor = None

    def __init__(self, **data):
        super().__init__(**data)
        self._openbabel_executor = OpenBabelExecutor()

    def execute(self):
        # compute either mcs or radial graph, end up with networkX object which needs to be parsed into a PerturbationMap object
        tmp_dir = self._make_tmpdir()
        existing_dir = os.getcwd()
        os.chdir(tmp_dir)
        self.write_conformers(os.path.join(tmp_dir, "confs.sdf"))
        # separate into individual files with obabel
        args = ["-isdf", "confs.sdf", "-osdf", "-O", "out.sdf", "-m"]
        self._openbabel_executor.execute(
            command=_OBE.OBABEL, arguments=args, check=True, location=tmp_dir
        )
        os.remove(os.path.join(tmp_dir, "confs.sdf"))

        # now rename each file such that it matches the index string
        for f in os.listdir(tmp_dir):
            with Chem.SDMolSupplier(f) as s:
                header = s[0].GetProp("_Name")
            os.rename(f, f"{header}.sdf")

        topology = self._get_additional_setting("topology")
        radial = True if topology == "radial" else False
        # hub_compound_name = self._get_additional_setting("hub_compound")
        # assert hub_compound_name is not None
        # hub_compound = self.get_compound_by_name(hub_compound_name)
        # if hub_compound:
        #     raise NameError(
        #         f"Hub compound {hub_compound_name} was not found in the compound list!"
        #     )
        # write mols out to sdf files

        dbmol = lomap.DBMolecules(
            tmp_dir, self.execution.parallelization.jobs, output=True, radial=radial
        )
        strict, loose = dbmol.build_matrices()
        strict_np = strict.to_numpy_2D_array()
        loose_np = loose.to_numpy_2D_array()

        # generate the nx graph
        nx_graph = dbmol.build_graph()

        p_map = PerturbationMap(
            compounds=self.get_compounds(),
        )
        p_map.generate_from_lomap_output(
            os.path.join(tmp_dir, "out_score_with_connection.txt")
        )
        # revert to old wd
        os.chdir(existing_dir)
        self.get_workflow_object().set_perturbation_map(p_map)



from typing import List, Union
from pydantic import BaseModel

from icolos.core.containers.compound import Compound, unroll_conformers
from icolos.core.step_utils.structcat_util import StructcatUtil
from icolos.core.step_utils.structconvert import StructConvert
from icolos.utils.enums.program_parameters import (
    OpenBabelEnum,
    SchrodingerExecutablesEnum,
)
from icolos.utils.enums.step_enums import (
    StepDataManipulationEnum,
    StepBaseEnum,
    StepFilterEnum,
)
from icolos.core.workflow_steps.io.base import StepIOBase
import os
from icolos.core.workflow_steps.step import _LE
import numpy as np
from rdkit import Chem
from rdkit.Chem import rdFMCS

_SBE = StepBaseEnum
_SDM = StepDataManipulationEnum()
_SEE = SchrodingerExecutablesEnum()
_OE = OpenBabelEnum()
_SFE = StepFilterEnum()


class StepDataManipulation(StepIOBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # extend parameters
        if _SDM.ACTION not in self.settings.additional.keys():
            self.settings.additional[
                _SDM.ACTION
            ] = _SDM.ACTION_ATTACH_CONFORMERS_AS_EXTRA
            self._logger.log(
                f"Action not specified, defaulting to {_SDM.ACTION_ATTACH_CONFORMERS_AS_EXTRA}.",
                _LE.WARNING,
            )

    def _attach_conformers_as_extra(self):
        # load data to match from previous step (note: no other input supported here, to avoid redudancy
        # with standard input preparation)
        match_compounds = (
            self.get_workflow_object()
            .find_step_by_step_id(self.settings.additional[_SDM.MATCH_SOURCE])
            .clone_compounds()
        )

        # unroll for convenience, attach matches to input conformers as extra data
        match_conformers = unroll_conformers(match_compounds)
        for comp in self.get_compounds():
            for enum in comp:
                for conf in enum:
                    list_matched = [
                        c
                        for c in match_conformers
                        if conf.get_index_string() == c.get_index_string()
                    ]
                    conf.add_extra_data(key=_SDM.KEY_MATCHED, data=list_matched)
                    self._logger.log(
                        f"Added {len(list_matched)} conformers as extra data to conformer {conf.get_index_string()}.",
                        _LE.DEBUG,
                    )

    def _convert_mae_to_pdb(self):
        converter = StructConvert(prefix_execution=_SEE.SCHRODINGER_MODULE)
        tmp_dir = self._make_tmpdir()

        # find the mae files from the input step and convert to pdb
        for file in self.data.generic.get_files_by_extension("mae"):
            file.write(tmp_dir)
            output_file = file.get_file_name().split(".")[0] + ".pdb"
            converter.mae2pdb(
                os.path.join(tmp_dir, file.get_file_name()),
                os.path.join(tmp_dir, output_file),
            )
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)

    def _assemble_complexes(self):
        concatenator = StructcatUtil(
            prefix_execution=_SEE.SCHRODINGER_MODULE, backend=_OE.OBABEL
        )
        assert os.path.isfile(self.settings.additional[_SDM.RECEPTOR])
        # create a tmpdir to work in
        tmp_dir = self._make_tmpdir()
        # get compounds from previous step
        conformers = self._unroll_compounds(self.get_compounds(), level="conformers")
        for conf in conformers:
            path = os.path.join(tmp_dir, f"{conf.get_index_string()}.sdf")
            conf.write(path)
            concatenator.concatenate(
                input_files=[
                    self.settings.additional[_SDM.RECEPTOR],
                    path,
                ],
                output_file=os.path.join(tmp_dir, f"{conf.get_index_string()}.pdb"),
            )
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)

    def _filter_compounds(self):
        """modifies set of input compounds according to the specification provided in the config block"""
        # TODO: support ranking structures based on generic data

        top_n = self.settings.additional[_SFE.RETURN_N]
        reverse = self.settings.additional[_SFE.HIGHEST_IS_BEST]
        criteria = (
            self.settings.additional[_SFE.CRITERIA]
            if _SFE.CRITERIA in self.settings.additional.keys()
            else None
        )
        aggregation = (
            self.settings.additional[_SFE.AGGREGATION]
            if _SFE.AGGREGATION in self.settings.additional.keys()
            else "sum"
        )

        top_conformer_list = []
        for compound in self.data.compounds:
            # filter by enumeration first - return a list of the top scoring conformers for that enumeration
            # this is the normal running mode, as opposed to sorting by compound, regardless of the enumeration it came from
            for enumeration in compound.get_enumerations():
                enumeration.sort_conformers(
                    by_tag=criteria, reverse=reverse, aggregation=aggregation
                )
                top_confs = enumeration.get_conformers()[:top_n]
                enumeration.clear_conformers()
                enumeration.add_conformers(top_confs)
                # replace that enumeration's conformers with the sorted list.
                # if filtering at conformer level i.e. regardless of enumeration
                if self.settings.additional[_SFE.FILTER_LEVEL] == _SFE.COMPOUNDS:
                    for conf in top_confs:
                        top_conformer_list.append(conf)
        if self.settings.additional[_SFE.FILTER_LEVEL] == _SFE.COMPOUNDS:
            self._logger.log("Filtering to top conformers per compound", _LE.DEBUG)

            # sort the top conformers from each enumeration and attach the top n conformers to their respective enumeration, get rid of the rest
            # sorted_top_confs = sorted(top_conformer_list,
            #                           key=lambda x: x.get_molecule().GetProp(self.settings.additional[_SFE.CRITERIA]),
            #                           reverse=reverse)[:top_n]
            # sort conformers
            sorted_top_confs = self._sort_conformers(
                conformers=top_conformer_list,
                by_tag=criteria,
                reverse=reverse,
                aggregation=aggregation,
            )

            for compound in self.data.compounds:
                for enum in compound.get_enumerations():
                    enum.clear_conformers()
            for conf in sorted_top_confs:
                enum = conf.get_enumeration_object()
                # if that enum's compound object already has top_n attached, skip
                comp: Compound = enum.get_compound_object()
                total_confs = self.get_conformer_count(comp)
                if total_confs < top_n:
                    enum.add_conformer(conf)
                else:
                    self._logger.log(
                        "Already reached max conformers, skipping", _LE.DEBUG
                    )

            # remove empty conformers

    def _compute_mcs(self):
        """Performs mcs calculation with a reference structure, attaches rmsd tag to conformers in place"""
        # modify compounds inplace

        suppl = Chem.SDMolSupplier(self.settings.additional["ref_lig"])

        ref = list(suppl)
        assert len(ref) == 1
        ref = ref[0]

        for comp in self.get_compounds():
            for enum in comp.get_enumerations():
                for conf in enum.get_conformers():
                    mol = conf.get_molecule()
                    r = rdFMCS.FindMCS([mol, ref])
                    a = mol.GetSubstructMatch(Chem.MolFromSmarts(r.smartsString))
                    b = ref.GetSubstructMatch(Chem.MolFromSmarts(r.smartsString))
                    atom_map = list(zip(a, b))
                    distances = []
                    for atomA, atomB in atom_map:
                        pos_A = mol.GetConformer().GetAtomPosition(atomA)
                        pos_B = ref.GetConformer().GetAtomPosition(atomB)
                        coord_A = np.array((pos_A.x, pos_A.y, pos_A.z))
                        coord_B = np.array((pos_B.x, pos_B.y, pos_B.z))
                        dist = np.linalg.norm(coord_A - coord_B)
                        distances.append(dist)
                    rmsd = np.sqrt(1 / len(distances) * sum([i * i for i in distances]))
                    mol.SetProp("rmsd", str(rmsd))
                    # these can be filtered in a second data manipulation step

    def get_conformer_count(self, comp: Compound) -> int:
        """count attached conformers over all enumerations

        :param Compound comp: compound to inspect
        :return int: number of conformers over all enumerations
        """
        total_confs = 0
        for enum in comp.get_enumerations():
            for conf in enum.get_conformers():
                total_confs += 1

        return total_confs

    def _sort_conformers(
        self,
        conformers,
        by_tag: Union[str, List[str]],
        reverse: bool = True,
        aggregation="sum",
    ):
        if isinstance(by_tag, list) and len(by_tag) == 1:
            by_tag = by_tag[0]

        if isinstance(by_tag, str):
            # sorting according to a single tag
            conformers = sorted(
                conformers,
                key=lambda x: float(x.get_molecule().GetProp(by_tag)),
                reverse=reverse,
            )
            return conformers
            # self._conformers = conformers
            # self.reset_conformer_ids()
        elif isinstance(by_tag, list):
            # need to normalise the values, calculate max and min of each tag for that series of conformers provided
            # this would allow us to compare across a series, i.e. scoring and ranking the output of all conformers in an enumeration from Glide
            def normalise_tag(value, tag):
                all_tag_values = [
                    float(conf.get_molecule().GetProp(tag)) for conf in conformers
                ]
                if len(all_tag_values) == 1:
                    return value
                else:

                    max_tag = np.max(all_tag_values)
                    min_tag = np.min(all_tag_values)
                    return (float(value) - min_tag) / (max_tag - min_tag)

            # if we specify multiple tags, aggregate according the the provided aggregation function
            if aggregation == "sum":
                # sort by the sum of the normalised tags,
                conformers = sorted(
                    conformers,
                    key=lambda x: np.sum(
                        [
                            float(normalise_tag(x.get_molecule().GetProp(i), i))
                            for i in by_tag
                        ]
                    ),
                    reverse=reverse,
                )
                return conformers
            elif aggregation == "product":
                conformers = sorted(
                    conformers,
                    key=lambda x: np.product(
                        [
                            float(normalise_tag(x.get_molecule().GetProp(i), i))
                            for i in by_tag
                        ]
                    ),
                    reverse=reverse,
                )
                return conformers
            else:
                raise AttributeError(
                    "Only sum or product aggregation modes are currently supported - ABORT"
                )

    def execute(self):
        if (
            self.settings.additional[_SDM.ACTION]
            == _SDM.ACTION_ATTACH_CONFORMERS_AS_EXTRA
        ):
            self._attach_conformers_as_extra()
        elif self.settings.additional[_SDM.ACTION] == _SDM.ACTION_NO_ACTION:
            n_comp, n_enum, n_conf = self.get_compound_stats()
            self._logger.log(
                f'Data manipulation step type "no_action" for {n_comp} compounds with {n_enum} enumerations with {n_conf} conformers completed.',
                _LE.INFO,
            )
        elif self.settings.additional[_SDM.ACTION] == _SDM.CONVERT_MAE_TO_PDB:
            self._convert_mae_to_pdb()
        elif self.settings.additional[_SDM.ACTION] == _SDM.ASSEMBLE_COMPLEXES:
            # take pose conformers (sd format) and concatenate with pdb file
            self._assemble_complexes()
        elif self.settings.additional[_SDM.ACTION] == _SDM.COLLECT_ITERATOR_RESULTS:
            # average the results coming from all iterations of the step
            raise NotImplementedError
        elif self.settings.additional[_SDM.ACTION] == _SDM.FILTER:
            self._filter_compounds()
            n_comp, n_enum, n_conf = self.get_compound_stats()
            self._logger.log(
                f"Filtered compounds, resulting in {n_comp} compounds with {n_enum} enumerations with {n_conf} conformers completed.",
                _LE.INFO,
            )
        elif self.settings.additional[_SDM.ACTION] == _SDM.COMPUTE_MCS:
            self._compute_mcs()

        else:
            raise ValueError(
                f'Action "{self.settings.additional[_SDM.ACTION]}" not supported.'
            )



from copy import deepcopy

import numpy as np
from typing import List

from pydantic import BaseModel

from icolos.core.containers.compound import Enumeration, Conformer

from icolos.utils.enums.step_enums import StepBoltzmannWeightingEnum
from icolos.core.workflow_steps.calculation.base import StepCalculationBase

from icolos.utils.general.convenience_functions import *
from icolos.utils.constants import *

_SBWE = StepBoltzmannWeightingEnum()


class StepBoltzmannWeighting(StepCalculationBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def _get_relative_energy_values(
        self, conformers: List[Conformer], property_name: str
    ) -> List[float]:
        values = [float(c.get_molecule().GetProp(property_name)) for c in conformers]
        min_val = min(values)
        relative_values = [value - min_val for value in values]
        return relative_values

    def _obtain_factors(self, relative_values: List[float]) -> List[float]:
        # calculate individual Boltzmann factors
        individual_factors = [
            np.exp((-1 * val / (CONSTANT_KB * CONSTANT_T))) for val in relative_values
        ]

        # calculate and return Boltzmann factors
        sum_factors = sum(individual_factors)
        factors = [val / sum_factors for val in individual_factors]
        return factors

    def _calculate_Boltzmann_factors(
        self, enumeration: Enumeration, parameters: dict
    ) -> List[str]:
        list_properties = parameters[_SBWE.PROPERTIES]
        list_output_names = []
        for prop in list_properties:
            # (1) get the relative values for this property (e.g. solvent) for all conformers in respect to the one
            # with the minimal energy
            relative_prop_values = self._get_relative_energy_values(
                conformers=enumeration.get_conformers(),
                property_name=prop[_SBWE.PROPERTIES_INPUT],
            )

            # (2) calculate the Boltzmann factors for this property
            boltzmann_factors = self._obtain_factors(
                relative_values=relative_prop_values
            )

            # (3) add the Boltzmann factors to the conformers as a tag
            for c, bm_factor in zip(enumeration.get_conformers(), boltzmann_factors):
                c.get_molecule().SetProp(prop[_SBWE.PROPERTIES_OUTPUT], str(bm_factor))
            list_output_names.append(prop[_SBWE.PROPERTIES_OUTPUT])
        return list_output_names

    def _do_Boltzmann_weighting(self, conformers: List[Conformer], weightings: dict):
        input_tags = weightings[_SBWE.WEIGHT_INPUT]
        output_prefix = nested_get(
            weightings, _SBWE.WEIGHT_OUTPUT_PREFIX, default="bf_weighted"
        )
        properties = weightings[_SBWE.WEIGHT_PROPERTIES]
        for prop in properties:
            for inp_tag in input_tags:
                new_tag_name = "_".join([output_prefix, inp_tag, prop])
                products = []
                for conformer in conformers:
                    conf = conformer.get_molecule()
                    products.append(
                        float(conf.GetProp(prop)) * float(conf.GetProp(inp_tag))
                    )
                for conformer in conformers:
                    conformer.get_molecule().SetProp(new_tag_name, str(sum(products)))

    def execute(self):
        parameters = deepcopy(self.settings.arguments.parameters)
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if self._input_object_empty(enumeration):
                    continue

                # get the name of the Boltzmann properties / solvents and annotate the factors
                _ = self._calculate_Boltzmann_factors(enumeration, parameters)

                # for each property and each weighting, add the respective tags
                self._do_Boltzmann_weighting(
                    conformers=enumeration.get_conformers(),
                    weightings=parameters[_SBWE.WEIGHT],
                )



from typing import List
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from icolos.core.workflow_steps.step import _LE
from pydantic import BaseModel
from icolos.utils.enums.program_parameters import (
    GromacsEnum,
)
from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.general.parallelization import SubtaskContainer
import os

_GE = GromacsEnum()


class StepPMXPrepareTransitions(StepPMXBase, BaseModel):
    """
    Prepare transitions: extract snapshots from equilibrium simulations, prepare .tpr files for each
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)

    def execute(self):
        edges = [e.get_edge_id() for e in self.get_edges()]

        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self.prepare_transitions,
            step_id="pmx prepare_transitions",
            result_checker=self._check_result,
        )

    def _extract_snapshots(self, eqpath, tipath):
        tpr = "{0}/tpr.tpr".format(eqpath)
        trr = "{0}/traj.trr".format(eqpath)
        frame = "{0}/frame.gro".format(tipath)

        trjconv_args = {
            "-s": tpr,
            "-f": trr,
            "-o": frame,
            "-sep": "",
            "-ur": "compact",
            "-pbc": "mol",
            "-b": 2000,
        }
        trjconv_args = self.get_arguments(trjconv_args)
        self._backend_executor.execute(
            _GE.TRJCONV, arguments=trjconv_args, pipe_input="echo System", check=False
        )

        last_frame = len([f for f in os.listdir(tipath) if f.startswith("frame")])
        self._logger.log(f"Extracted {last_frame} frames", _LE.DEBUG)

        self._clean_backup_files(tipath)
        # once frames are extracted, remove the large trr file from the equilibrium run
        files_to_remove = [trr, tpr, os.path.join(eqpath, "frame.gro")]
        for f in files_to_remove:
            try:
                os.remove(f)
            except FileNotFoundError:
                pass

    def _prepare_system(self, edge: str, state: str, wp: str, r: int, toppath: str):
        eqpath = self._get_specific_path(
            workPath=self.work_dir,
            edge=edge,
            wp=wp,
            state=state,
            r=r,
            sim="eq",
        )
        tipath = self._get_specific_path(
            workPath=self.work_dir,
            edge=edge,
            wp=wp,
            state=state,
            r=r,
            sim="transitions",
        )
        # if the trr file exists and snapshots have not been extracted in a previous run
        if os.path.isfile(os.path.join(eqpath, "traj.trr")) and not os.path.isfile(
            os.path.join(tipath, "frame0.gro")
        ):
            self._extract_snapshots(eqpath, tipath)
        else:
            self._logger.log("Skipping frame extraction, already present", _LE.DEBUG)
        print("preparing tpr files")

        self._prepare_single_tpr(
            simpath=tipath,
            toppath=toppath,
            state=state,
            sim_type="transitions",
            executor=self._backend_executor,
        )

        self._clean_backup_files(tipath)

    def prepare_transitions(self, jobs: List[str]):
        for edge in jobs:
            ligTopPath = self._get_specific_path(
                workPath=self.work_dir, edge=edge, wp="unbound"
            )
            protTopPath = self._get_specific_path(
                workPath=self.work_dir, edge=edge, wp="bound"
            )
            for state in self.states:
                for r in range(1, self.get_perturbation_map().replicas + 1):

                    self._logger.log(
                        f"Preparing transitions: {edge}, {state}, run {r}", _LE.DEBUG
                    )
                    self._prepare_system(
                        edge=edge, state=state, wp="unbound", r=r, toppath=ligTopPath
                    )
                    self._prepare_system(
                        edge=edge, state=state, wp="bound", r=r, toppath=protTopPath
                    )

    def _check_result(self, batch: List[List[str]]) -> List[List[bool]]:
        """
        Look in each hybridStrTop dir and check the output pdb files exist for the edges
        """
        replicas = self.get_perturbation_map().replicas
        output_paths = []
        for i in range(1, replicas + 1):
            output_paths.append(f"unbound/stateA/run{i}/transitions")
            output_paths.append(f"unbound/stateB/run{i}/transitions")
            output_paths.append(f"bound/stateA/run{i}/transitions")
            output_paths.append(f"bound/stateB/run{i}/transitions")

        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                # check number of frames extracted == number tpr files for each leg

                num_frames = [
                    len(
                        [
                            f
                            for f in os.listdir(os.path.join(self.work_dir, job, f))
                            if f.startswith("frame")
                        ]
                    )
                    for f in output_paths
                ]
                num_tprs = [
                    len(
                        [
                            f
                            for f in os.listdir(os.path.join(self.work_dir, job, f))
                            if f.startswith("ti")
                        ]
                    )
                    for f in output_paths
                ]
                self._logger.log(
                    f"Found {num_tprs} tpr files and {num_frames} frame files for edge {job}",
                    _LE.DEBUG,
                )
                # confirms that frames have been extracted, and we have a tpr file generated for each ti frame
                subjob_results.append(
                    num_tprs == num_frames and all(n > 10 for n in num_frames)
                )
            results.append(subjob_results)
        return results
